"""
=================
Creating a Reader
=================

**Suhas Somnath**

8/28/2020

This document illustrates an example of extracting data out of proprietary raw
data files, thereby describing how one would write a ``sidpy.Reader`` class.

The captured information would be populated into a / set of ``sidpy.Dataset``
object(s) as appropriate.

Introduction
------------
In most scientific disciplines, commercial instruments tend to write the data and metadata out into proprietary file
formats that significantly impede access to the data and metadata, thwart sharing of data and correlation of data from
multiple instruments, and complicate long-term archival, among other things. One of the data wrangling steps in science
is the extraction of the data and metadata out of the proprietary file formats and writing the information into files
that are easier to access, share, etc. The overwhelming part of this data wrangling effort is in investigating how to
extract the data and metadata into memory. Often, the data and parameters in these files are **not** straightforward to
access. In certain cases, additional / dedicated software packages are necessary to access the data while in many other
cases, it is possible to extract the necessary information from built-in **numpy** or similar python packages included
with **anaconda**. Once the information is accessible in the computer memory, such as in the
form of numpy arrays, scientists have a wide variety of tools to write the data out into files.

Simpler data such as images or single spectra can easily be written into plain text files. Simple or complex / large /
multidimensional data can certainly be stored as numpy data files. However, there are significant drawbacks to writing
data into non-standardized structures or file formats. First, while the structure of the data and metadata may be
intuitive for the original author of the data, that may not be the case for another researcher. Furthermore, such
formatting may change from a day-to-day basis. As a consequence, it becomes challenging to develop code that can accept
such data whose format keeps changing.

One solution to these challenges is to write the data out into standardized files such as ``h5USID`` files.
The USID model aims to make data access, storage, curation, etc. simply by storing the data along with all
relevant parameters in a single file (HDF5 for now).

The process of copying data from the original format to **h5USID** files is called
**Translation** and the classes available in pyUSID and children packages such as pycroscopy that perform these
operation are called **Translators**.

As we alluded to earlier, the process of developing a ``sidpy.Reader`` can be
broken down into two basic components:

1. Extracting data and metadata out of the proprietary file format
2. Populating one or more ``sidpy.Dataset`` objects as necessary

This process is the same regardless of the origin, complexity, or size of the scientific data. It is not necessary that
the two components be disjoint - there are many situations where both components may need to happen simultaneously
especially when the data sizes are very large.

The goal of this document is to demonstrate how one would extract data and parameters from a Scanning Tunnelling
Spectroscopy (STS) raw data file obtained from an Omicron Scanning Tunneling Microscope (STM).
In this dataset, a spectra was collected for each position in a two-dimensional grid of spatial locations, thereby
resulting in a 3D dataset.

The code in this example is an abbreviation of the
`AscTranslator <https://github.com/pycroscopy/pycroscopy/blob/master/pycroscopy/io/translators/omicron_asc.py>`_
available in our sister package - `pycroscopy`.

Recommended pre-requisite reading
=================================

Before proceeding with this example, we recommend learning about ``Sidpy.Reader``:

.. tip::
    You can download and run this document as a Jupyter notebook using the link at the bottom of this page.

Import all necessary packages
=============================
There are a few setup procedures that need to be followed before any code is written. In this step, we simply load a
few python packages that will be necessary in the later steps.
"""

# Ensure python 3 compatibility:
from __future__ import division, print_function, absolute_import, unicode_literals

# The package for accessing files in directories, etc.:
import os
import zipfile

# Warning package in case something goes wrong
from warnings import warn
import subprocess
import sys


def install(package):
    subprocess.call([sys.executable, "-m", "pip", "install", package])
# Package for downloading online files:
try:
    # This package is not part of anaconda and may need to be installed.
    import wget
except ImportError:
    warn('wget not found.  Will install with pip.')
    import pip
    install(wget)
    import wget

# The mathematical computation package:
import numpy as np

# The package used for creating and manipulating HDF5 files:
import h5py

# Packages for plotting:
import matplotlib.pyplot as plt

# import sidpy - supporting package for pyUSID:
try:
    import sidpy
except ImportError:
    warn('sidpy not found.  Will install with pip.')
    import pip
    install('sidpy')
    import sidpy

####################################################################################
# Procure the Raw Data file
# =========================
# Here we will download a compressed data file from Github and unpack it:

url = 'https://raw.githubusercontent.com/pycroscopy/pyUSID/master/data/STS.zip'
zip_path = 'STS.zip'
if os.path.exists(zip_path):
    os.remove(zip_path)
_ = wget.download(url, zip_path, bar=None)

zip_path = os.path.abspath(zip_path)
# figure out the folder to unzip the zip file to
folder_path, _ = os.path.split(zip_path)
zip_ref = zipfile.ZipFile(zip_path, 'r')
# unzip the file
zip_ref.extractall(folder_path)
zip_ref.close()
# delete the zip file
os.remove(zip_path)

data_file_path = 'STS.asc'

####################################################################################
# 1. Extracting data and metadata from proprietary files
# ------------------------------------------------------
# 1.1 Explore the raw data file
# =============================
#
# Inherently, one may not know how to read these ``.asc`` files. One option is to try and read the file as a text file
# one line at a time.
#
# If one is lucky, as in the case of these ``.asc`` files, the file can be read like conventional text files.
#
# Here is how we tested to see if the ``asc`` files could be interpreted as text files. Below, we read just the first 10
# lines in the file

with open(data_file_path, 'r') as file_handle:
    for lin_ind in range(10):
        print(file_handle.readline().replace('\n', ''))

####################################################################################
# 1.2 Read the contents of the file
# =================================
# Now that we know that these files are simple text files, we can manually go through the file to find out which lines
# are important, at what lines the data starts etc.
# Manual investigation of such ``.asc`` files revealed that these files are always formatted in the same way. Also, they
# contain instrument- and experiment-related parameters in the first ``403`` lines and then contain data which is
# arranged as one pixel per row.
#
# STS experiments result in 3 dimensional datasets ``(X, Y, current)``. In other words, a 1D array of current data (as a
# function of excitation bias) is sampled at every location on a two dimensional grid of points on the sample.
# By knowing where the parameters are located and how the data is structured, it is possible to extract the necessary
# information from these files.
#
# Since we know that the data sizes (<200 MB) are much smaller than the physical memory of most computers, we can start
# by safely loading the contents of the entire file to memory.

# Reading the entire file into memory
with open(data_file_path, 'r') as file_handle:
    string_lines = file_handle.readlines()

####################################################################################
# 1.3 Extract the metadata
# ========================
# In the case of these ``.asc`` files, the parameters are present in the first few lines of the file. Below we will
# demonstrate how we parse the first 17 lines to extract some very important parameters. Note that there are several
# other important parameters in the next 350 or so lines. However, in the interest of brevity, we will focus only on the
# first few lines of the file. The interested reader is recommended to read the ``ASCTranslator`` available in
# ``pycroscopy`` for more complete details.

# Preparing an empty dictionary to store the metadata / parameters as key-value pairs
parm_dict = dict()

# Reading parameters stored in the first few rows of the file
for line in string_lines[3:17]:
    # Remove the hash / pound symbol, if any
    line = line.replace('# ', '')
    # Remove new-line escape-character, if any
    line = line.replace('\n', '')
    # Break the line into two parts - the parameter name and the corresponding value
    temp = line.split('=')
    # Remove spaces in the value. Remember, the value is still a string and not a number
    test = temp[1].strip()
    # Now, attempt to convert the value to a number (floating point):
    try:
        test = float(test)
        # In certain cases, the number is actually an integer, check and convert if it is:
        if test % 1 == 0:
            test = int(test)
    except ValueError:
        pass
    parm_dict[temp[0].strip()] = test

# Print out the parameters extracted
for key in parm_dict.keys():
    print(key, ':\t', parm_dict[key])

####################################################################################
# At this point, we recommend reformatting the parameter names to standardized nomenclature.
# We realize that the materials imaging community has not yet agreed upon standardized nomenclature for metadata.
# Therefore, we leave this as an optional, yet recommended step.
# For example, in pycroscopy, we may categorize the number of rows and columns in an image under ``grid`` and
# data sampling parameters under ``IO``.
# As an example, we may rename ``x-pixels`` to ``positions_num_cols`` and ``y-pixels`` to ``positions_num_rows``.
#
# 1.4 Extract parameters that define dimensions
# =============================================
# Just having the metadata above and the main measurement data is insufficient to fully describe experimental data.
# We also need to know how the experimental parameters were varied to acquire the multidimensional dataset at hand.
# In other words, we need to answer how the grid of locations was defined and how the bias was varied to acquire the
# current information at each location. This is precisely what we will do below.
#
# Since, we did not parse the entire list of parameters present in the file above, we will need to make some up.
# Please refer to the formal ``ASCTranslator`` to see how this step would have been different.

num_rows = int(parm_dict['y-pixels'])
num_cols = int(parm_dict['x-pixels'])
num_pos = num_rows * num_cols
spectra_length = int(parm_dict['z-points'])

# We will assume that data was collected from -3 nm to +7 nm on the Y-axis or along the rows
y_vec = np.linspace(-3, 7, num_rows, endpoint=True)

# We will assume that data was collected from -5 nm to +5 nm on the X-axis or along the columns
x_vec = np.linspace(-5, 5, num_cols, endpoint=True)

# The bias was sampled from -1 to +1 V in the experiment. Here is how we generate the Bias axis:
bias_vec = np.linspace(-1, 1, spectra_length)

####################################################################################
# 1.5 Extract the data
# ====================
# We have observed that the data in these ``.asc`` files are consistently present after the first ``403`` lines of
# parameters. Using this knowledge, we need to populate a data array using data that is currently present as text lines
# in memory (from step 2).
#
# These ``.asc`` file store the 3D data (X, Y, spectra) as a 2D matrix (positions, spectra). In other words, the spectra
# are arranged one below another. Thus, reading the 2D matrix from top to bottom, the data arranged column-by-column,
# and then row-by-row So, for simplicity, we will prepare an empty 2D numpy array to store the data as it exists in the
# raw data file.
#
# Recall that in step 2, we were lucky enough to read the entire data file into memory given its small size.
# The data is already present in memory as a list of strings that need to be parsed as a matrix of numbers.

num_headers = 403

raw_data_2d = np.zeros(shape=(num_pos, spectra_length), dtype=np.float32)

# Iterate over ever measurement position:
for pos_index in range(num_pos):
    # First, get the correct (string) line corresponding to the current measurement position.
    # Recall that we would need to skip the many header lines to get to the data
    this_line = string_lines[num_headers + pos_index]
    # Each (string) line contains numbers separated by tabs (``\t``). Let us break the line into several shorter strings
    # each containing one number. We will ignore the last entry since it is empty.
    string_spectrum = this_line.split('\t')[:-1]  # omitting the new line
    # Now that we have a list of numbers represented as strings, we need to convert this list to a 1D numpy array
    # the converted array is set to the appropriate position in the main 2D array.
    raw_data_2d[pos_index] = np.array(string_spectrum, dtype=np.float32)

####################################################################################
# If the data is so large that it cannot fit into memory, we would need to read data one (or a few) position(s) at a
# time, process it (e.g. convert from string to numbers), and write it to the HDF5 file without keeping much or any data
# in memory.
#
# The three-dimensional dataset (``Y``, ``X``, ``Bias``) is currently represented as a two-dimensional array:
# (``X`` * ``Y``, ``Bias``). To make it easier for us to understand and visualize, we can turn it into a 3D array:

raw_data_3d = raw_data_2d.reshape(num_rows, num_cols, spectra_length)
print('Shape of 2D data: {}, Shape of 3D data: {}'.format(raw_data_2d.shape, raw_data_3d.shape))

####################################################################################
# Just as we did for the parameters (``X``, ``Y``, and ``Bias``) that were varied in the experiment,
# we need to specify the quantity that is recorded from the sensors / detectors, units, and what the data
# represents:

main_data_name = 'STS'
main_qty = 'Current'
main_units = 'nA'

####################################################################################
# Visualize the extracted data
# ============================
# Here is a visualization of the current-voltage spectra at a few locations:

fig, axes = sidpy.plot_utils.plot_curves(bias_vec, raw_data_2d, num_plots=9,
                                        x_label='bias (V)',
                                        y_label=main_qty + '(' + main_units + ')',
                                        title='Current-Voltage Spectra at different locations',
                                        fig_title_yoffset=1.05)

####################################################################################
# Here is a visualization of spatial maps at different bias values

fig, axes = sidpy.plot_utils.plot_map_stack(raw_data_3d, reverse_dims=True, pad_mult=(0.15, 0.15),
                                           title='Spatial maps of current at different bias', stdevs=2,
                                           color_bar_mode='single', num_ticks=3, x_vec=x_vec, y_vec=y_vec,
                                           evenly_spaced=True, fig_mult=(3, 3), title_yoffset=0.95)

for axis, bias_ind in zip(axes, np.linspace(0, len(bias_vec), 9, endpoint=False, dtype=np.uint)):
    axis.set_title('Bias = %3.2f V' % bias_vec[bias_ind])

####################################################################################
# 2. Populating the ``Dataset`` object
# ====================================
data_set = sidpy.Dataset.from_array(raw_data_3d, name='Raw_Data')
print(data_set)

####################################################################################
# Dimensions
data_set.set_dimension(0, sidpy.Dimension('y', y_vec,units='nm',
                                          quantity='Length',
                                          dimension_type='spatial'))
data_set.set_dimension(1, sidpy.Dimension('x', x_vec, units='nm',
                                          quantity='Length',
                                          dimension_type='spatial'))
data_set.set_dimension(2, sidpy.Dimension('bias', bias_vec,
                                          quantity='Bias',
                                          dimension_type='spectral'))

####################################################################################
# Generic top level metadata can be added as you go along
data_set.data_type = 'Current'
data_set.units = main_units
data_set.quantity = main_qty

####################################################################################
# Instrument-specific metadata
data_set.metadata = parm_dict

####################################################################################
# Print the dataset object
print(data_set)

####################################################################################
# Visualize the dataset object
data_set.plot()

####################################################################################
# How does one turn such ad-hoc code into a ``Reader`` class?

####################################################################################
# More information
# ================
# Our sister class - pycroscopy, has several
# `translators <https://github.com/pycroscopy/pycroscopy/tree/master/pycroscopy/io/translators>`_ that translate popular
# file formats generated by nanoscale imaging instruments.
# These will be moved to SciFiReaders soon
#
# We have found python packages online to open a few proprietary file formats and have written translators using these
# packages. If you are having trouble reading the data in your files and cannot find any packages online, consider
# contacting the manufacturer of the instrument which generated the data in the proprietary format for help.
#
# Cleaning up
# ===========
# Remove the original file:
os.remove(data_file_path)



# -*- coding: utf-8 -*-
"""
Created on Fri Nov 5 16:43:00 2021

@author: Rama Vasudevan
"""

import numpy as np  # For array operations
import sidpy as sid
from sidpy.sid import Reader, Dimension, DimensionType
from .nanonis_base import Grid
class Nanonis3dsReader(Reader):

    def __init__(self, file_path, *args, **kwargs):

        super().__init__(file_path, *args, **kwargs)

    @staticmethod
    def _parse_3ds_parms(header_dict, signal_dict):
        """
        Parse 3ds files.
        Parameters
        ----------
        header_dict : dict
        signal_dict : dict
        Returns
        -------
        parm_dict : dict
        """
        parm_dict = dict()
        data_dict = dict()

        # Create dictionary with measurement parameters
        meas_parms = {key: value for key, value in header_dict.items()
                      if value is not None}
        channels = meas_parms.pop('channels')
        for key, parm_grid in zip(meas_parms.pop('fixed_parameters')
                                  + meas_parms.pop('experimental_parameters'),
                                  signal_dict['params'].T):
            # Collapse the parm_grid along one axis if it's constant
            # along said axis
            if parm_grid.ndim > 1:
                dim_slice = list()
                # Find dimensions that are constant
                for idim in range(parm_grid.ndim):
                    tmp_grid = np.moveaxis(parm_grid.copy(), idim, 0)
                    if np.all(np.equal(tmp_grid[0], tmp_grid[1])):
                        dim_slice.append(0)
                    else:
                        dim_slice.append(slice(None))
                # print(key, dim_slice)
                # print(parm_grid[tuple(dim_slice)])
                parm_grid = parm_grid[tuple(dim_slice)]
            meas_parms[key] = parm_grid
        parm_dict['meas_parms'] = meas_parms

        # Create dictionary with channel parameters and
        # save channel data before renaming keys
        data_channel_parms = dict()
        for chan_name in channels:
            splitted_chan_name = chan_name.split(maxsplit=2)
            if len(splitted_chan_name) == 2:
                direction = 'forward'
            elif len(splitted_chan_name) == 3:
                direction = 'backward'
                splitted_chan_name.pop(1)
            name, unit = splitted_chan_name
            key = ' '.join((name, direction))
            data_channel_parms[key] = {'Name': name,
                                       'Direction': direction,
                                       'Unit': unit.strip('()'),
                                       }
            data_dict[key] = signal_dict.pop(chan_name)
        parm_dict['channel_parms'] = data_channel_parms

        # Add remaining signal_dict elements to data_dict
        data_dict.update(signal_dict)

        # Position dimensions
        nx, ny = header_dict['dim_px']
        if 'X (m)' in parm_dict:
            row_vals = parm_dict.pop('X (m)')
        else:
            row_vals = np.arange(nx, dtype=np.float32)

        if 'Y (m)' in parm_dict:
            col_vals = parm_dict.pop('Y (m)')
        else:
            col_vals = np.arange(ny, dtype=np.float32)
        pos_vals = np.hstack([row_vals.reshape(-1, 1),
                              col_vals.reshape(-1, 1)])
        pos_names = ['X', 'Y']

        dims = [Dimension(values, name=label, quantity='Length', units='nm',
                          dimension_type=DimensionType.SPATIAL)
                for label, values in zip(pos_names, pos_vals.T)]

        # Spectroscopic dimensions
        sweep_signal = header_dict['sweep_signal']
        spec_label, spec_unit = sweep_signal.split(maxsplit=1)
        spec_unit = spec_unit.strip('()')
        # parm_dict['sweep_signal'] = (sweep_name, sweep_unit)
        dc_offset = data_dict['sweep_signal']
        spec_dim = Dimension(dc_offset, quantity='Bias', name=spec_label,
                             units=spec_unit,
                             dimension_type=DimensionType.SPECTRAL)
        dims.append(spec_dim)
        data_dict['Dimensions'] = dims

        return parm_dict, data_dict

    def read(self):
        """
        Returns
        -------
        list of sidpy.Dataset objects containing the spectroscopy data
        """
        
        reader = Grid
        override_header = {
            'Delay before measuring (s)': 0.0,
            'Start time': 0.0,
            'End time': 1000.0,
            'Comment': 'Default values for delay before measuring (s), Start time and End time fields were used! Beware!'
        }
        nanonis_data = reader(self._input_file_path, header_override=override_header)

        header_dict = nanonis_data.header
        signal_dict = nanonis_data.signals

        parm_dict, data_dict = self._parse_3ds_parms(header_dict,
                                                         signal_dict)
      
        self.parm_dict = parm_dict
        self.data_dict = data_dict

        #Specify dimensions
        x_dim = self.data_dict['Dimensions'][0]
        y_dim = self.data_dict['Dimensions'][1]
        z_dim = self.data_dict['Dimensions'][2]

        dataset_list = []
        channel_parms = self.parm_dict['channel_parms']
        orig_metadata = self.parm_dict['meas_parms']

        chan_names = list(self.parm_dict['channel_parms'].keys())

        for dataset_name in chan_names:
            
            data_mat = self.data_dict[dataset_name]
            
            #Make a sidpy dataset
            data_set = sid.Dataset.from_array(data_mat, name = dataset_name)

            #Set the data type
            data_set.data_type = sid.DataType.SPECTRAL_IMAGE
            
            metadata = channel_parms[dataset_name]

            # Add quantity and units
            data_set.units = metadata['Unit']
            data_set.quantity = metadata['Name']

            # Add dimension info
            data_set.set_dimension(0, x_dim)
            data_set.set_dimension(1, y_dim)
            data_set.set_dimension(2, z_dim)
        
            # append metadata 
            def merge_dict(dict1, dict2):
                res = {**dict1, **dict2}
                return res
            
            chan_metadata = self.parm_dict['channel_parms'][dataset_name]
            
            data_set.original_metadata =  merge_dict(chan_metadata,orig_metadata)
            dataset_list.append(data_set)
        
        return dataset_list
        
        return 

    def can_read(self):
        """
        Tests whether or not the provided file has a .3ds extension
        Returns
        -------
        """

        return 



"""
################################################################################
# Python class for reading Bruker .rto files into sidpy Dataset
# and extracting all metadata
#
# Written by Gerd Duscher, UTK 2025
#
# Works for python 3
#
################################################################################
"""
import os
import codecs
import xml

import numpy as np
import sidpy

__version__ = '0.1beta'
__all__ = ["BrukerReader", "__version__"]


def get_bruker_dictionary(filename):
    """ Convert xm, style .rto file to python dictionary

    The problem is that there are xml sections with same name (spectrum, image) that need to be numbered
    """
    tree = xml.etree.ElementTree.parse(filename)
    root = tree.getroot()
    spectrum_number = 0
    i = 0
    image_count = 0
    overlay_count = 0
    tags = {}
    for neighbor in root.iter():
        if 'Type' in neighbor.attrib:
            if 'TRTCrossOverlayElement' in neighbor.attrib['Type']:
                if 'Spectrum' in neighbor.attrib['Name']:
                    overlay_count += 1
                    if 'overlay' not in tags:
                        tags['overlay'] = {}
                    if 'image' + str(image_count) not in tags['overlay']:
                        tags['overlay']['image' + str(image_count)] = {}
                    tags['overlay']['image' + str(image_count)][neighbor.attrib['Name']] = {}

                    over = tags['overlay']['image' + str(image_count)][neighbor.attrib['Name']]

                    for child in neighbor.iter():
                        if 'verlay' in child.tag:
                            pos = child.find('./Pos')
                            if pos is not None:
                                over['pos_x'] = int(pos.find('./PosX').text)
                                over['pos_y'] = int(pos.find('./PosY').text)
                                over['position'] = [over['pos_x'], over['pos_y']]
                                over['type'] = 'spot'
                                over['label'] = neighbor.attrib['Name']
            if 'TRTImageData' in neighbor.attrib['Type']:
                _ = neighbor.find("./ClassInstance[@Type='TRTCrossOverlayElement']")
                image = neighbor
                if 'image' not in tags:
                    tags['image'] = {}
                tags['image'][image_count] = {}
                im = tags['image'][image_count]
                im['width'] = int(image.find('./Width').text)  # in pixels
                im['height'] = int(image.find('./Height').text)  # in pixels
                im['dtype'] = 'u' + image.find('./ItemSize').text  # in bytes ('u1','u2','u4')
                im['scale_x'] = float(image.find('./XCalibration').text)
                im['scale_y'] = float(image.find('./YCalibration').text)
                im['plane_count'] = int(image.find('./PlaneCount').text)
                im['date'] = str((image.find('./Date').text))
                im['time'] = str((image.find('./Time').text))
                im['data'] = {}
                for j in range(im['plane_count']):
                    img = image.find("./Plane" + str(i))
                    raw = codecs.decode((img.find('./Data').text).encode('ascii'), 'base64')
                    array1 = np.frombuffer(raw, dtype=im['dtype'])
                    im['data'][str(j)] = np.reshape(array1, (im['height'], im['width']))
                image_count += 1
            if 'TRTDetectorHeader' == neighbor.attrib['Type']:
                detector = neighbor
                tags['detector'] = {}
                for child in detector:
                    if child.tag == "WindowLayers":
                        tags['detector']['layers'] = {}
                        for child2 in child:
                            tags['detector']['layers'][child2.tag] = {}
                            tags['detector']['layers'][child2.tag]['Z'] = child2.attrib["Atom"]
                            tags['detector']['layers'][child2.tag]['thickness'] = float(
                                child2.attrib["Thickness"]) * 1e-6  # now in m
                            if 'RelativeArea' in child2.attrib:
                                tags['detector']['layers'][child2.tag]['relative_area'] = float(
                                    child2.attrib["RelativeArea"])
                    else:
                        if child.tag != 'ResponseFunction':
                            if child.text is not None:
                                tags['detector'][child.tag] = child.text

                                if child.tag == 'SiDeadLayerThickness':
                                    tags['detector'][child.tag] = float(child.text) * 1e-6

                                if child.tag == 'DetectorThickness':
                                    tags['detector'][child.tag] = float(child.text) * 1e-1

            # ESMA could stand for Electron Scanning Microscope Analysis
            if 'TRTESMAHeader' == neighbor.attrib['Type']:
                esma = neighbor
                tags['esma'] = {}
                for child in esma:
                    if child.tag in ['PrimaryEnergy', 'ElevationAngle', 'AzimutAngle', 'Magnification',
                                     'WorkingDistance']:
                        tags['esma'][child.tag] = float(child.text)
            if 'TRTSpectrum' == neighbor.attrib['Type']:
                if 'spectrum' not in tags:
                    tags['spectrum'] = {}
                if 'Name' in neighbor.attrib:
                    spectrum = neighbor
                    trt_header = spectrum.find('./TRTHeaderedClass')

                    if trt_header is not None:
                        hardware_header = trt_header.find("./ClassInstance[@Type='TRTSpectrumHardwareHeader']")
                        spectrum_header = spectrum.find("./ClassInstance[@Type='TRTSpectrumHeader']")
                        result_header = spectrum.find("./ClassInstance[@Type='TRTResult']")
                        tags['spectrum'][spectrum_number] = {}
                        tags['spectrum'][spectrum_number]['hardware_header'] = {}
                        if hardware_header is not None:
                            for child in hardware_header:
                                tags['spectrum'][spectrum_number]['hardware_header'][child.tag] = child.text
                        tags['spectrum'][spectrum_number]['detector_header'] = {}
                        tags['spectrum'][spectrum_number]['spectrum_header'] = {}
                        for child in spectrum_header:
                            tags['spectrum'][spectrum_number]['spectrum_header'][child.tag] = child.text
                        tags['spectrum'][spectrum_number]['results'] = {}
                        for result in result_header:
                            result_tag = {}
                            for child in result:
                                result_tag[child.tag] = child.text
                            if 'Atom' in result_tag:
                                if result_tag['Atom'] not in tags['spectrum'][spectrum_number]['results']:
                                    tags['spectrum'][spectrum_number]['results'][result_tag['Atom']] = {}
                                tags['spectrum'][spectrum_number]['results'][result_tag['Atom']].update(result_tag)
                        tags['spectrum'][spectrum_number]['data'] = np.frombuffer(spectrum.find('./Channels').text,
                                                                                  dtype='np.int16', sep=",")
                        spectrum_number += 1
    return tags


def get_image(tags, key=0):
    """ Convert image part of dictionary to sidpy dataset"""
    image = tags['image'][key]
    names = ['x', 'y']
    units = 'nm'
    quantity = 'distance'
    dimension_type = 'spatial'
    to_nm = 1e3

    scale_x = float(image['scale_x']) * to_nm
    scale_y = float(image['scale_y']) * to_nm

    dataset = sidpy.Dataset.from_array(image['data']['0'].T)
    dataset.data_type = 'image'
    dataset.units = 'counts'
    dataset.quantity = 'intensity'

    dataset.modality = 'SEM'
    dataset.title = 'image'
    dataset.add_provenance('SciFiReader', 'BrukerReader', version='0', linked_data='File: ')

    dataset.set_dimension(0, sidpy.Dimension(np.arange(image['data']['0'].shape[1]) * scale_x,
                                             name=names[0], units=units,
                                             quantity=quantity,
                                             dimension_type=dimension_type))
    dataset.set_dimension(1, sidpy.Dimension(np.arange(image['data']['0'].shape[0]) * scale_y,
                                             name=names[1], units=units,
                                             quantity=quantity,
                                             dimension_type=dimension_type))

    dataset.metadata['experiment'] = tags.get('esma', {}).copy()
    dataset.metadata['annotations'] = tags.get('overlay', {}).get('image1', {}).copy()
    return dataset


def get_spectrum(tags, key=0):
    """ Convert spectrum part of dictionary to sipdy dataset"""
    spectrum = tags['spectrum'][key]
    offset = float(spectrum['spectrum_header']['CalibAbs'])
    scale = float(spectrum['spectrum_header']['CalibLin'])

    dataset = sidpy.Dataset.from_array(spectrum['data'])
    energy_scale = (np.arange(len(dataset)) * scale + offset) * 1000
    dataset.units = 'counts'
    dataset.quantity = 'intensity'
    dataset.data_type = 'spectrum'

    dataset.modality = 'EDS'
    dataset.title = 'spectrum'
    dataset.set_dimension(0, sidpy.Dimension(energy_scale,
                                             name='energy_scale', units='eV',
                                             quantity='energy',
                                             dimension_type='spectral'))

    dataset.metadata['experiment'] = tags['esma'].copy()
    acceleration_voltage = float(tags.get('experiment', {}).get('PrimaryEnergy', 0))*1000.
    dataset.metadata['experiment']['acceleration_voltage'] =  acceleration_voltage
    dataset.metadata['EDS'] = {'detector': tags.get('detector', '')}
    dataset.metadata['EDS']['results'] = {}
    for  result in spectrum.get('results', {}).values():
        if 'Name' in result:
            dataset.metadata['EDS']['results'][result['Name']] = result.copy()
    return dataset


class BrukerReader(sidpy.Reader):
    """
    Creates an instance of BrukerReader which can read .rto files
    datasets formatted in the Bruker format

    We can read Images, and Spectra.
    Please note that all original metadata are retained in each sidpy dataset.
    ToDo: evaluate if this should this be a point cloud

    Parameters
    ----------
    file_path : str
        Path to a Bruker rto file
    Return
    ------
    datasets: dict
        dictionary of sidpy.Datasets
    """

    def __init__(self, file_path, verbose=False):
        """
        file_path: filepath to rto file.
        """

        super().__init__(file_path)

        # initialize variables ##
        self.verbose = verbose
        self.__filename = file_path

        _, file_name = os.path.split(self.__filename)
        self.basename, self.extension = os.path.splitext(file_name)
        self.datasets = {}
        self.tags = {}
        if 'rto' in self.extension:
            try:
                self.tags = get_bruker_dictionary(file_path)

            except IOError:
                raise IOError(f"File {self.__filename} does not seem to be of Bruker's .rto format")
        self.channel_number = 0

    def can_read(self):
        """ Checks if the reader can read the file format """
        return 'rto' in self.extension

    def read(self):
        """ Parses the dictionary that was extracted from the xml type file of Bruker .rto file
        Reads images and spectra
        """
        if 'image' in self.tags:
            key = f"Channel_{int(self.channel_number):03d}"
            self.datasets[key] = get_image(self.tags, key=0)
            self.channel_number += 1

            self.datasets[key].title = self.basename+'_image'
            self.datasets[key].add_provenance('SciFiReader', 'BrukerReader', version=__version__,
                                              linked_data=f'File: {self.basename}.{self.extension}')
        if 'spectrum' in self.tags:
            for spectrum_key in self.tags['spectrum']:
                key = f"Channel_{int(self.channel_number):03d}"
                self.datasets[key] = get_spectrum(self.tags, key=spectrum_key)
                self.channel_number += 1

                self.datasets[key].title = self.basename + '_spectrum_' + str(spectrum_key)
                self.datasets[key].add_provenance('SciFiReader', 'BrukerReader', version=__version__,
                                                  linked_data=f'File: {self.basename}.{self.extension}')
        return self.datasets

    def get_filename(self):
        """ Returns the filename of the Bruker file"""
        return self.__filename

    def get_datasets(self):
        """ Returns the datasets created by the reader""" 
        return self.datasets

    filename = property(get_filename)



# -*- coding: utf-8 -*-
"""
Created on Wed Dec 07 16:04:34 2016

Updated on Fri Sep 11 15:22:00 2020 for ScopeReader

@author: Suhas Somnath, Chris R. Smith, Raj Giridhargopal, Rama Vasudevan
"""

from __future__ import division, print_function, absolute_import, unicode_literals
import sys
import numpy as np  # For array operations
import sidpy as sid
from sidpy.sid import Reader
import h5py
import os
import re

if sys.version_info.major == 3:
    unicode = str

# Reader Information
# ----------------------
reader_name = 'ARHDF5'
description = 'Asylum Research HDF5 file typically used for storing spatial maps of force-distance curves'

class ARhdf5Reader(Reader):
    """
    Extracts data and metadata from ARhdf5 files
    These are Asylum Research files typically captured for
    force maps from their microscopes.
    The ARhdf5 file should be generated with the converter provided
    by Asylum Research called ARDFtoHDF5. Contact David Aue <David.Aue@oxinst.com>
    or Tommaso Costanzo <tommaso.costanzo01@gmail.com> to get a
    copy of the converter. NOTE: the AR converter works only under
    windows.

    NOTE: At this point, the z dimension vector is not correctly captured
    Not immediately clear how to fix it. TODO: Check with Asylum
    """

    def read(self, verbose=False):
        """
        Reads the file given in file_path into a sidpy dataset

        Parameters
        ----------
        verbose : Boolean (Optional)
            Whether or not to show  print statements for debugging
        Returns
        -------
        sidpy.Dataset : List of sidpy.Dataset objects.
            Multi-channel inputs are separated into individual dataset objects
        """
        file_path = self._input_file_path

        self.notes = None
        self.segments = None
        self.segments_name = []
        self.map_size = {'X': 0, 'Y': 0}
        self.channels_name = []
        self.points_per_sec = None
        self.verbose = verbose

        # Open the datafile
        try:
            data_filepath = os.path.abspath(file_path)
            ARh5_file = h5py.File(data_filepath, 'r')
        except:
            print('Unable to open the file', data_filepath)
            raise

        # Get info from the origin file like Notes and Segments
        try:
            self.notes = ARh5_file.attrs['Note'].decode('utf-8')
        except AttributeError:
            self.notes = ARh5_file.attrs['Note']

        self.segments = ARh5_file['ForceMap']['0']['Segments']
        segments_name = list(ARh5_file['ForceMap']['0'].attrs['Segments'])
        try:
            self.segments_name = [name.decode('utf-8') for name in segments_name]
        except AttributeError:
            self.segments_name = [name for name in segments_name]

        self.map_size['X'] = ARh5_file['ForceMap']['0']['Segments'].shape[0]
        self.map_size['Y'] = ARh5_file['ForceMap']['0']['Segments'].shape[1]
        channels_name = list(ARh5_file['ForceMap']['0'].attrs['Channels'])
        
        try:
            self.channels_name = [name.decode('utf-8') for name in channels_name]
        except AttributeError:
            self.channels_name = [name for name in channels_name]

        try:
            self.points_per_sec = float(self.note_value('ARDoIVPointsPerSec'))
        except NameError:
            self.points_per_sec = float(self.note_value('NumPtsPerSec'))

        if self.verbose:
            print('Map size [X, Y]: ', self.map_size)
            print('Channels names: ', self.channels_name)
            print('Name of Segments:', self.segments_name)

        # Only the extension 'Ext' segment can change size
        # so we get the shortest one and we trim all the others

        extension_idx = self.segments_name.index('Ext')
        short_ext = np.amin(np.array(self.segments[:, :, extension_idx]))
        longest_ext = np.amax(np.array(self.segments[:, :, extension_idx]))
        difference = longest_ext - short_ext  # this is a difference between integers
        tot_length = (np.amax(self.segments) - difference) + 1
        # +1 otherwise array(tot_length) will be of 1 position shorter
        points_trimmed = np.array(self.segments[:, :, extension_idx]) - short_ext

        # Open the output hdf5 file
        x_dim = np.linspace(0, float(self.note_value('FastScanSize')),
                            self.map_size['X'])
        y_dim = np.linspace(0, float(self.note_value('FastScanSize')),
                            self.map_size['Y'])
        z_dim = np.arange(tot_length) / float(self.points_per_sec)

        datasets = [] #list of sidpy datasets

        #Let's get all the metadata
        # Create the new segments that will be stored as attribute
        new_segments = {}
        for seg, name in enumerate(self.segments_name):
            new_segments.update({name: self.segments[0, 0, seg] - short_ext})

        seg_metadata = {'Segments': new_segments,
                                           'Points_trimmed': points_trimmed,
                                           'Notes': self.notes}
        general_metadata = {'translator': 'ARhdf5',
                            'instrument': 'Asylum Research ' + self.note_value('MicroscopeModel'),
                            'AR sftware version': self.note_value('Version')}

        combined_metadata = dict({'segments': seg_metadata, 'general': general_metadata})

        for index, channel in enumerate(self.channels_name):
            main_dset = np.empty((self.map_size['X'], self.map_size['Y'], tot_length))
            for column in np.arange(self.map_size['X']):
                for row in np.arange(self.map_size['Y']):
                    AR_pos_string = str(column) + ':' + str(row)
                    seg_start = self.segments[column, row, extension_idx] - short_ext
                    main_dset[column, row, :] = ARh5_file['ForceMap']['0'][AR_pos_string][index, seg_start:]

            quant_unit = self.get_def_unit(channel)

            data_set = sid.Dataset.from_array(main_dset, name='Image')
            data_set.data_type = sid.DataType.SPECTRAL_IMAGE

            # Add quantity and units
            data_set.units = quant_unit
            data_set.quantity = channel

            # Add dimension info
            data_set.set_dimension(0, sid.Dimension(x_dim,
                                                    name='x',
                                                    units='m', quantity='x',
                                                    dimension_type='spatial'))
            data_set.set_dimension(1, sid.Dimension(y_dim,
                                                    name='y',
                                                    units='m', quantity='y',
                                                    dimension_type='spatial'))
            data_set.set_dimension(2, sid.Dimension(z_dim,
                                                    name='Time',
                                                    units='s', quantity='z',
                                                    dimension_type='spectral'))

            # append metadata
            chan_metadata = dict(ARh5_file['ForceMap']['0'].attrs)
            new_dict = {**chan_metadata, **combined_metadata}
            data_set.metadata = new_dict

            # Finally, append it
            datasets.append(data_set)

        for index, image in enumerate(ARh5_file['Image'].keys()):
            main_dset = np.array(ARh5_file['Image'][image])

            quant_unit = self.get_def_unit(image)

            data_set = sid.Dataset.from_array(main_dset, name='Image')
            data_set.data_type = sid.DataType.IMAGE

            # Add quantity and units
            data_set.units = quant_unit
            data_set.quantity = channel

            # Add dimension info
            data_set.set_dimension(0, sid.Dimension(x_dim,
                                                    name='x',
                                                    units=quant_unit, quantity='x',
                                                    dimension_type='spatial'))
            data_set.set_dimension(1, sid.Dimension(y_dim,
                                                    name='y',
                                                    units=quant_unit, quantity='y',
                                                    dimension_type='spatial'))

            # append metadata
            chan_metadata = dict(ARh5_file['Image'].attrs)
            new_dict = {**chan_metadata, **combined_metadata}
            data_set.metadata = new_dict

            # Finally, append it
            datasets.append(data_set)

        # Return the dataset
        return datasets

    def can_read(self):
        """
        Tests whether or not the provided file has a .ibw extension
        Returns
        -------

        """
        file_path = self._input_file_path
        extension = os.path.splitext(file_path)[1][1:]
        if extension not in ['hdf5', 'h5']:
            return False
        try:
            h5_f = h5py.File(file_path, 'r')
        except:
            return False
        if 'ForceMap' not in h5_f.keys():
            return False
        if not isinstance(h5_f['ForceMap'], h5py.Group):
            return False
        return

    def note_value(self, name):
        '''
        Get the value of a single note entry with name "name"

        Parameters
        ----------------
        name : String / unicode
            Name of the parameter to get the value

        Returns
        ----------------
        value : String / unicode
            Value of the Note entry requested.
        '''
        try:
            match = re.search(r"^" + name + r":\s+(.+$)", self.notes, re.M)
            if not match:
                raise Exception
        except:
            match = re.search(r"^" + name + ":+(.+$)", self.notes, re.M)
        if (match):
            matched = match.groups()
            if len(matched) == 1:
                return match.groups()[0]
            else:
                # We do not expect to enter here
                print('WARNING! Multiple value matched! \n Only the first is returned')
                return match.groups()[0]
        else:
            raise NameError('Note entry with name "{}" not found'.format(name))

    def get_def_unit(self, chan_name):
        """
        Retrive the default unit from the channel name

        Parameters
        ----------
        chan_name : string
            Name of the channel to get the unit

        Returns
        -------
        default_unit : string
            Default unit of that channel
        """

        # Check if chan_name is string
        if not isinstance(chan_name, (str, unicode)):
            raise TypeError('The channel name must be of type string')

        # Find the default unit
        if chan_name.startswith('Phas'):
            default_unit = 'deg'
        elif chan_name.startswith('Curr'):
            default_unit = 'A'
        elif chan_name.startswith('Freq'):
            default_unit = 'Hz'
        elif chan_name.startswith('Bias'):
            default_unit = 'V'
        elif (chan_name.startswith('Amp') or
              chan_name.startswith('Raw') or
              chan_name.startswith('ZSnsr') or
              chan_name.startswith('Defl') or
              chan_name.startswith('MapHeight')):
            default_unit = 'm'
        elif (chan_name.startswith('Seconds') or
              chan_name == 'TriggerTime'):
            default_unit = 's'
        elif chan_name.startswith('HeaterTemperature'):
            default_unit = 'Celsius'
        elif chan_name == 'MapAdhesion':
            default_unit = 'N/m^2'
        elif chan_name == 'HeaterHumidity':
            default_unit = 'g/m^3'
        elif chan_name.endswith('LVDT'):
            # This should be the laser virtual deflection
            default_unit = 'm'
        else:
            if self.verbose:
                print('Unknown unit for channel: {}'.format(chan_name))
                print('Unit set to "unknown"')
            default_unit = 'unknown'

        return default_unit



"""
Tools to read, data in from TEM files

Submodules
----------

.. autosummary::
    :toctree: _autosummary

    dm_reader
    nion_reader
"""

from .edax_reader import EDAXReader
from .bruker_reader import BrukerReader

__all__ = ['EDAXReader', 'BrukerReader']

all_readers = [EDAXReader, BrukerReader]


################################################################################
# Python class for reading FEI Velox 4D STEM .mrc files into sidpy Dataset
# and extracting all metadata
#
# Written by Austin Houston, UTK 2024
#
# based on MRC2014 file format:
# https://www.ccpem.ac.uk/mrc_format/mrc2014.php
#
# and corresponding mrcfile package:
# https://pypi.org/project/mrcfile/
################################################################################


import json
import h5py
import sys
import numpy as np
import dask.array as da
import sidpy
import mrcfile
try:
    from tqdm.auto import tqdm
    tqdm_available = True
except ImportError:
    tqdm_available = False

__all__ = ["MRCReader", "version"]

version = '1.0'


class MRCReader(sidpy.Reader):

    def __init__(self, file_path):
        super(MRCReader, self).__init__(file_path)
        self.file_path = file_path
        self.metadata = None
        self.data = None
        self.scan_shape = None

        self.dataset = None

            
    def read(self, handedness='right', scan_pixel_size=None):
        # scan pixel size needs to be in meters, if you know it
        # handedness is either 'right' or 'left', and determines the handedness of the scan
        # not all .mrc files have the handedness embedded, so if your file looks wrong, try changing this

        # read with 
        mrc =  mrcfile.mmap(self.file_path, permissive=True)

        # data
        mrc_data = mrc.data

        # metadata
        extended_header = mrc.indexed_extended_header
        metadata_labels = extended_header.dtype.names
        metadata_labels = [label for label in metadata_labels]

        # Reshape the data
        scan_shape_params = ['Scan size right', 'Scan size left', 'Scan size top', 'Scan size bottom']

        sizes = []
        for label in scan_shape_params:
            size = np.unique(extended_header[label])
            sizes.append(size)
        sizes = np.array(sizes).flatten()
        y_shape = int(np.abs(sizes[0] - sizes[1]))
        x_shape = int(np.abs(sizes[2] - sizes[3]))
        reshape_target = (x_shape, y_shape, mrc_data.shape[-2], mrc_data.shape[-1])

        try:
            self.data = np.reshape(mrc_data, reshape_target)
        except ValueError:
            print(f'Error reshaping data: {mrc_data.shape} to {reshape_target}')
            print(f'the scan must have been stopped early, on the microscope - this creates issues still')
            print(f'sorry, we do not support reading point cloud versions of this data yet')


        # These 'pixel sizes' are usually in the order of 10^8: This the camera pixel size, not the scan step size.
        # I've talked to thermofisher and plan to update this eventually (2024-9-6)
        camera_pixel_sizes = []
        for label in ['Pixel size X', 'Pixel size Y']:
            size = np.unique(extended_header[label])
            size *= 1e-10 # conversion from 1/m to 1/Angstrom
            camera_pixel_sizes.append(size)

        # scan pixel size
        if scan_pixel_size:
            # if scan pixel size is given, use it
            self.scan_pixel_size = scan_pixel_size * 1e10 # conversion from m to Angstrom
            self.scan_size_units = 'Å'
        else:
            self.scan_pixel_size = 1
            self.scan_size_units = 'pixels'


        # make metadata dictionary
        metadata = {}
        for label in metadata_labels:
            # later, we may want to remove 'np.unique()', but I see no problems now
            metadata[label] = np.unique(extended_header[label])
        self.metadata = metadata


        # create sidpy Dataset
        dataset = sidpy.Dataset.from_array(self.data, name='MRC_000', chunks=(1, 1, reshape_target[-2], reshape_target[-1]))
        # add metadata dictionary
        dataset.original_metadata = self.metadata
        dataset.data_type = 'image_4d'

        dataset.set_dimension(0, sidpy.Dimension(np.arange(dataset.shape[0]) * self.scan_pixel_size, 
                                                name='x', units=self.scan_size_units, quantity='length',
                                                dimension_type='spatial'))

        dataset.set_dimension(1, sidpy.Dimension(np.arange(dataset.shape[1]) * self.scan_pixel_size,
                                                name='y', units=self.scan_size_units, quantity='length',
                                                dimension_type='spatial'))

        dataset.set_dimension(2, sidpy.Dimension(np.arange(dataset.shape[2]) * camera_pixel_sizes[0],
                                                name='u', units='1/Å', quantity='angle',
                                                dimension_type='reciprocal'))

        dataset.set_dimension(3, sidpy.Dimension(np.arange(dataset.shape[3]) * camera_pixel_sizes[1],
                                                name='v', units='1/Å', quantity='angle',
                                                dimension_type='reciprocal'))

        return {'Channel_000': dataset}
    




""" 
################################################################################
# Python class for reading FEI Velox .emd files into sidpy Dataset
# and extracting all metadata
#
# Written by Gerd Duscher, UTK 2021
# update 08/2025 - improved EDS data reading
#
################################################################################
"""
import json
import h5py
import numpy as np
# import dask.array as da
from numba import njit
import sidpy

__version__ = '0.1beta'
__all__ = ["EMDReader", "__version__"]

DEBUG_LEVEL = 0  # 0=none, 1-3=basic, 4-5=simple, 6-10 verbose


class EMDReader(sidpy.Reader):
    """
    Creates an instance of EMDReader which can read one or more HDF5
    datasets formatted in the FEI Velox style EDM file

    We can read Images, and SpectrumStreams (SpectrumImages and Spectra).
    Please note that all original metadata are retained in each sidpy dataset.

    Parameters
    ----------
    file_path : str
        Path to a HDF5 file
    Return
    ------
    datasets: dict
        dictionary of sidpy.Datasets
    """
    def __init__(self, file_path:str, sum_frames:bool=False, no_eds:bool=False):
        """
        Initialize an EMDReader instance.

        Parameters:
            file_path (str): Path to the HDF5 (.emd) file to be read.
            sum_frames (bool, optional): If True, sum all frames in the dataset. Defaults to False.
            no_eds (bool, optional): If True, disables EDS (Energy Dispersive Spectroscopy) 
            data handling. Defaults to False.

        Attributes:
            _h5_file (h5py.File): The opened HDF5 file object.
            datasets (dict): Dictionary to store dataset references.
            channel_number (int): The current channel number (default is 0).
            key (str): Formatted string key for the current channel.
            data_array (ndarray or None): Array to hold data from the file.
            metadata (dict or None): Metadata extracted from the file.
            label_dict (dict): Dictionary for label mapping.
            no_eds (bool): Indicates if EDS data handling is disabled.
            sum_frames (bool): Indicates if frames should be summed.
            number_of_frames (int): Number of frames in the dataset (default is 1).
        """
        super().__init__(file_path)

        # Let h5py raise an OS error if a non-HDF5 file was provided
        self._h5_file: h5py.File = h5py.File(file_path, mode='r+')

        self.datasets = {}
        self.channel_number = 0
        self.key = f"Channel_{int(self.channel_number):03d}"
        self.data_array = None
        self.metadata = None
        self.label_dict = {}
        self.no_eds = no_eds
        self.sum_frames = sum_frames
        self.number_of_frames = 1
        self.image_key = ''
        self.bin_xy = 1

    def can_read(self):
        """
        Checks whether or not this Reader can read the provided file

        Returns
        -------
        bool :
            True if this Reader can read the provided file
            Else, False
        """
        if 'Application' in self._h5_file:
            return 'Velox' in self._h5_file['Application']
        return False

    def read(self, eds_stream:bool=False, bin_xy:int=2):
        """
        Reads all available datasets in FEI Velox style hdf5 files with .edm

        Parameters
        ----------
        eds_stream: boolean
            switch to return spectrum image (default - False) or original spectrum stream (True)
        bin_xy: int
            binning factor for EDS spectrum size reduction

        Returns
        -------
        datasets: list of sidpy.Dataset objects
            Datasets present in the provided file
        """

        if 'Data' not in self._h5_file:
            raise TypeError('Velox EMD File is empty')

        number_of_datasets = 0

        self.bin_xy = bin_xy
        self.image_key = ''

        for key in self._h5_file['Data'].keys():  # type: str
            if key == 'SpectrumStream':
                number_of_datasets += len(self._h5_file['Data']['SpectrumStream'].keys())
        for key in self._h5_file['Data'].keys():
            self.image_key = 'None'
            self._parse_image_display()
            if key == 'Image':
                for self.image_key in self._h5_file['Data']['Image']:
                    self.get_data('Data/Image/' + self.image_key)
                    self.get_image()
                    self.extract_crucial_metadata(self.key)
            elif key == 'SpectrumStream':
                if not self.no_eds:
                    for stream_key in self._h5_file['Data']['SpectrumStream']:
                        self.get_data('Data/SpectrumStream/' + stream_key)
                        self.get_eds(eds_stream)
                        self.extract_crucial_metadata(self.key)
        self.close()
        return self.datasets

    def get_data(self, image_key):
        """Get the image data and metadata from the file."""
        self.data_array = self._h5_file[image_key]['Data']
        metadata_array = self._h5_file[image_key]['Metadata'][:, 0]
        metadata_string = metadata_array.tobytes().decode("utf-8")
        self.metadata = dict(json.loads(metadata_string.rstrip('\x00')))
        if 'AcquisitionSettings' in self._h5_file[image_key]:
            self.metadata['AcquisitionSettings'] = json.loads(self._h5_file[image_key]['AcquisitionSettings'][0])

    def _parse_image_display(self):
        ### Read image label according to hyperspy EMDReader
        image_display_group = self._h5_file.get('Presentation/Displays/ImageDisplay')
        if image_display_group is not None:
            for key in image_display_group.keys():
                v = json.loads(
                    image_display_group[key][0].decode('utf-8'))
                data_key = v['dataPath'].split('/')[-1]  # key in data group
                self.label_dict[data_key] = v['display']['label']
        else:
            image_display_group = self._h5_file.get('Displays/ImageDisplay')
            if image_display_group is not None:
                for key in image_display_group.keys():
                    v = json.loads(
                        image_display_group[key][0].decode('utf-8'))
                    if 'data' in v:
                        data = json.loads(self._h5_file[v['data']][()][0].decode('utf-8'))
                        data_key = data['dataPath'].split('/')[-1]  # key in data group
                        self.label_dict[data_key] = v['title']


    def get_eds(self, eds_stream=False):
        """Get the EDS data from the file."""
        if 'AcquisitionSettings' not in self.metadata:
            eds_stream = True
        key = f"Channel_{int(self.channel_number):03d}"
        self.key = key
        self.channel_number += 1
        if eds_stream:
            self.datasets[key] = sidpy.Dataset.from_array(self.data_array)
        else:
            data_array = self.get_eds_spectrum()
            if data_array.shape[0] == 1 and data_array.shape[1] == 1:
                data_array = np.squeeze(data_array)
                chunks = 1
            else:
                chunks= [data_array.shape[1], 32, data_array.shape[2]]
                if data_array.shape[0]> chunks[0]:
                    chunks[0] = data_array.shape[0]
                if data_array.shape[1]> chunks[1]:
                    chunks[1] = data_array.shape[1]
            self.datasets[key] = sidpy.Dataset.from_array(data_array, chunks=chunks)

        self.data_array=np.zeros([1,1])
        self.datasets[key].original_metadata = self.metadata
        detectors = self.datasets[key].original_metadata.get('Detectors', {})

        print(detectors)
        if eds_stream:
            pass
        else:
            offset = 0
            dispersion = 1.
            for detector in detectors.values():
                if self.metadata.get('BinaryResult', {}).get('Detector') in detector.get('DetectorName', ''):
                    offset = float(detector.get('OffsetEnergy', 0))
                    dispersion = float(detector.get('Dispersion', 0))*self.bin_xy*1.003
            self.datasets[key].units = 'counts'
            self.datasets[key].quantity = 'intensity'
            self.datasets[key].modality = 'EDS'

            energy_scale = np.arange(self.datasets[key].shape[-1]) * dispersion + offset

            if self.datasets[key].ndim == 1:
                self.datasets[key].data_type = 'spectrum'

                self.datasets[key].set_dimension(0, sidpy.Dimension(energy_scale,
                                                                   name='energy_scale', units='eV',
                                                                   quantity='energy',
                                                                   dimension_type='spectral'))

            else:
                self.datasets[key].data_type = 'spectral_image'
                print(self.datasets[key].shape)

                scale =  self.metadata.get('BinaryResult', {}).get('PixelSize', {})
                scale_x = float(scale.get('width', 1e-9)) * 1e9
                scale_y = float(scale.get('height', 1e-9)) * 1e9

                self.datasets[key].set_dimension(0, sidpy.Dimension(
                                            np.arange(self.datasets[key].shape[0]) * scale_x,
                                            name='x', units='nm',
                                            quantity='distance',
                                            dimension_type='spatial'))
                self.datasets[key].set_dimension(1, sidpy.Dimension(np.arange(
                                            self.datasets[key].shape[1]) * scale_y,
                                            name='y', units='nm',
                                            quantity='distance',
                                            dimension_type='spatial'))
                self.datasets[key].set_dimension(2, sidpy.Dimension(
                                            energy_scale,
                                            name='energy_scale', units='eV',
                                            quantity='energy',
                                            dimension_type='spectral'))

    def get_eds_spectrum(self) -> np.ndarray:
        """Get the EDS spectrum."""
        acquisition = self.metadata.get('AcquisitionSettings', {})

        scan_area = self.metadata.get('Scan', {}).get('ScanArea', {})
        scan_size = self.metadata.get('Scan', {}).get('ScanSize', {})

        if len(scan_area) * len(scan_size) > 0:
            size_x = float(scan_size.get('width', 1)) * float(scan_area.get('right', 1))
            size_x -= float(scan_size.get('width', 1)) * float(scan_area.get('left', 0))
            size_y = float(scan_size.get('height', 1)) * float(scan_area.get('bottom', 1))
            size_y -= float(scan_size.get('height', 1)) * float(scan_area.get('top', 0))
            size_x = int(size_x)
            size_y = int(size_y)
        else:
            raster_scan = acquisition.get('RasterScanDefinition', {})
            size_x = int(raster_scan.get('Width', 1))
            size_y = int(raster_scan.get('Height', 1))

        spectrum_size = int(acquisition.get('bincount', 0))
        self.number_of_frames = np.ceil((self.data_array[:, 0] == 65535).sum())
        self.number_of_frames = int(self.number_of_frames / (size_x * size_y))

        data_array = np.zeros((int(size_x*size_y), int(spectrum_size/self.bin_xy)),dtype=np.ushort)

        data, frame = get_stream(data_array, size_x*size_y, self.data_array[:, 0], self.bin_xy)

        self.number_of_frames = frame
        return np.reshape(data, (size_x, size_y, int(spectrum_size/self.bin_xy)))

    def get_image(self):
        """ Get the image data."""
        key = f"Channel_{int(self.channel_number):03d}"
        self.key = key
        self.channel_number += 1

        if self.metadata['BinaryResult']['PixelUnitX'] == '1/m':
            names = ['u', 'v']
            units = '1/nm'
            quantity = 'reciprocal distance'
            dimension_type='reciprocal'
            to_nm = 1e-9
        else:
            names = ['x', 'y']
            units = 'nm'
            quantity = 'distance'
            dimension_type='spatial'
            to_nm = 1e9
        scale = self.metadata.get('BinaryResult', {})
        scale_x = float(scale.get('PixelSize', {}).get('width', 0)) * to_nm
        scale_y = float(scale.get('PixelSize', {}).get('height', 0)) * to_nm
        offset_x = float(scale.get('Offset', {}).get('x', 0)) * to_nm
        offset_y = float(scale.get('Offset', {}).get('y', 0)) * to_nm

        if self.sum_frames:
            data_array = np.zeros([self.data_array.shape[0], self.data_array.shape[1], 1])
            for i in range(self.data_array.shape[2]):
                data_array[:, :, 0] += self.data_array[:, :, i]
            self.data_array = data_array

        if self.data_array.shape[2] == 1:
            self.datasets[key] = sidpy.Dataset.from_array(self.data_array[:, :, 0])
            self.datasets[key].data_type = 'image'
            self.datasets[key].set_dimension(0, sidpy.Dimension(
                                    np.arange(self.data_array.shape[0]) * scale_x + offset_x,
                                    name=names[0], units=units,
                                    quantity=quantity,
                                    dimension_type=dimension_type))
            self.datasets[key].set_dimension(1, sidpy.Dimension(
                                    np.arange(self.data_array.shape[1]) * scale_y + offset_y,
                                    name=names[1], units=units,
                                    quantity=quantity,
                                    dimension_type=dimension_type))
        else:
            # There is a problem with random access of data due to chunking in hdf5 files
            # Speed-up copied from hyperspy.ioplugins.EMDReader.FEIEMDReader
            data_array = np.empty(self.data_array.shape)
            self.data_array.read_direct(data_array)
            self.data_array = np.rollaxis(data_array, axis=2)

            self.datasets[key] = sidpy.Dataset.from_array(self.data_array)
            self.datasets[key].data_type = 'image_stack'

            self.datasets[key].set_dimension(0, sidpy.Dimension(
                                    np.arange(self.data_array.shape[0]),
                                    name='frame', units='frame',
                                    quantity='time',
                                    dimension_type='temporal'))
            self.datasets[key].set_dimension(1, sidpy.Dimension(
                                    np.arange(self.data_array.shape[1]) * scale_x + offset_x,
                                    name=names[0], units=units,
                                    quantity=quantity,
                                    dimension_type=dimension_type))
            self.datasets[key].set_dimension(2, sidpy.Dimension(
                                    np.arange(self.data_array.shape[2]) * scale_y + offset_y,
                                    name=names[1], units=units,
                                    quantity=quantity,
                                    dimension_type=dimension_type))
        self.datasets[key].original_metadata = self.metadata

        self.datasets[key].units = 'counts'
        self.datasets[key].quantity = 'intensity'
        self.datasets[key].modality = 'image'

        if self.image_key in self.label_dict:
            self.datasets[key].title = self.label_dict[self.image_key]
        self.data_array=np.zeros([1,1])

    def extract_crucial_metadata(self, key):
        """ Extract some crucial metadata from the original metadata."""
        metadata = self.datasets[key].original_metadata
        experiment = {'detector': metadata.get('BinaryResult', {}).get('Detector', ''),
                      'acceleration_voltage': float(metadata.get('Optics', {}).get('AccelerationVoltage', 0)),
                      'microscope': metadata.get('Instrument', {}).get('InstrumentClass', ''),
                      'start_date_time': int(metadata.get('Acquisition', {}).get('AcquisitionStartDatetime', {}).get('DateTime', 0)),
                      'collection_angle': 0.0,
                      'convergence_angle': 0.0}
        if metadata.get('Optics', {}).get('ProbeMode') == "1":
            experiment['probe_mode'] = "convergent"
            experiment['convergence_angle'] = float(metadata.get('Optics', {}).get('BeamConvergence', 0))
        else:  # metadata['Optics']['ProbeMode'] == "2":
            experiment['probe_mode'] = "parallel"
        stage_dict = metadata.get('Stage', {})
        experiment['stage'] = {"holder": "",
                               "position": {"x": float(stage_dict.get('Position', {}).get('x', 0)),
                                            "y": float(stage_dict.get('Position', {}).get('y', 0)),
                                            "z": float(stage_dict.get('Position', {}).get('z', 0))},
                                "tilt": {"alpha": float(stage_dict.get('AlphaTilt', 0)),
                                         "beta": float(stage_dict.get('BetaTilt', 0))}}
        model = self.datasets[key].original_metadata.get('Instrument', {}).get('InstrumentModel', '')
        instrument_id = self.datasets[key].original_metadata.get('Instrument', {}).get('InstrumentId', '')
        experiment['instrument'] = model + str(instrument_id)
        experiment['current'] = float(metadata.get('Optics', {}).get('LastMeasuredScreenCurrent', 0.0))
        experiment['pixel_time'] = float(metadata.get('Scan', {}).get('DwellTime', 0.0))
        experiment['exposure_time'] = float(metadata.get('Scan', {}).get('FrameTime', 0.0))

        if isinstance(metadata.get('Sample'), dict):
            experiment['sample'] = metadata['Sample'].get('SampleDescription', '')
            experiment['sample_id'] = metadata['Sample'].get('SampleId', '')
        eds = {}
        used_detector = experiment.get('detector', 'unknown')
        for detector in metadata.get('Detectors', {}).values():
            if used_detector in detector.get('DetectorName', ''):
                begin = detector.get('CollectionAngleRange', {}).get('begin', -1)
                experiment['collection_angle'] = float(begin)
                end = detector.get('CollectionAngleRange', {}).get('end', -1)
                experiment['collection_angle_end'] = float(end)
                if 'SuperX' in detector['DetectorName']:
                    start_energy = int(detector.get('BeginEnergy', 120))
                    energy_scale = self.datasets[key].get_spectral_dims(return_axis=True)
                    if len(energy_scale) >0:
                        start_channel = 100
                    else:
                        start_channel = np.searchsorted(energy_scale, start_energy)
                    eds = {'detector': {'layers': {13: {'thickness': 0.05*1e-6, 'Z': 13, 'element': 'Al'}},
                                        'SiDeadThickness': .13 *1e-6,  # in m
                                        'SiLiveThickness': 0.05 , # in m
                                        'detector_area': 30 * 1e-6, #in m2
                                        'energy_resolution': 125,  # in eV
                                        'start_energy': start_energy,  # in eV
                                        'start_channel': int(start_channel),
                                        'ElevationAngle': float(detector.get('ElevationAngle', 0.0)), 
                                        'AzimuthAngle': float(detector.get('AzimuthAngle', 0.0)),
                                        'RealTime': float(detector.get('RealTime', 0.0)), 
                                        'LiveTime': float(detector.get('LiveTime', 0.0))}}
        self.datasets[key].metadata['experiment'] = experiment.copy()
        if eds:
            self.datasets[key].metadata['EDS'] = eds.copy()

        if self.datasets[key].title == 'generic':
            self.datasets[key].title = experiment['detector']

    def close(self):
        """ Close the h5 file."""
        self._h5_file.close()

@njit(cache=True)
def get_stream(data, size, data_stream, bin_xy):
    """ Get the EDS spectrum from the stream."""
    #for value in self.data_array[:, 0]:
    #from tqdm.auto import trange, tqdm
    pixel_number = 0
    frame = 0
    for value in data_stream:
        if value == 65535:
            pixel_number += 1
            if pixel_number >= size:
                pixel_number = 0
                frame += 1
        else:
            data[pixel_number, int(value/bin_xy-0.2)] += 1
    return data, frame



from . import microscopy, SID, generic
from .microscopy import *
from .SID import *
from .generic import *
from .spectroscopy import *
from .converters import *

__all__ = microscopy.__all__ + generic.__all__ + SID.__all__ + \
          spectroscopy.__all__ + converters.__all__

all_readers = microscopy.all_readers + generic.all_readers + \
              SID.all_readers + spectroscopy.all_readers   #+ \
              # converters.all_readers



# -*- coding: utf-8 -*-
"""
Created on Wed Dec 07 16:04:34 2016

Updated on Fri Sep 11 15:22:00 2020 for ScopeReader

@author: Suhas Somnath, Chris R. Smith, Raj Giridhargopal, Rama Vasudevan
"""

import sys
import numpy as np  # For array operations
import sidpy as sid
from sidpy.sid import Reader
import re
from warnings import warn

try:
    from igor2 import binarywave as bw
except ModuleNotFoundError:
    warn("You don't have igor2 installed. \
    If you wish to open igor files, you will need to install it \
    (pip install igor2) before attempting.", ImportWarning)
    bw = None

class IgorMatrixReader(Reader):
    """
    Extracts data and metadata from Igor Binary Wave (.ibw) files exported from MatrixFiles
    This is a very specific reader that cannot be reconciled with the generic ibw reader due to severe inconsistencies.
    Unless you exported Matrix files as ibw files using IGOR5 Exporter v120401, this reader will not work.
    """
    def __init__(self, file_path, *args, **kwargs):
        if bw == None:
            raise ModuleNotFoundError('You attempted to load an Igor file, but this requires igor2.\n \
            Please Load it with pip install igor2 , restart and retry')

        super().__init__(file_path, *args, **kwargs)

    def read(self, verbose=False, parm_encoding='utf-8'):
        """
        Reads the file given in file_path into a sidpy dataset

        Parameters
        ----------
        verbose : Boolean (Optional)
            Whether or not to show  print statements for debugging
        parm_encoding : str, optional
            Codec to be used to decode the bytestrings into Python strings if
            needed. Default 'utf-8'

        Returns
        -------
        sidpy.Dataset : List of sidpy.Dataset objects.
            Multi-channel inputs are separated into individual dataset objects
        """
        file_path = self._input_file_path

        # Load the ibw file first
        ibw_obj = bw.load(file_path)
        ibw_wave = ibw_obj.get('wave')
        parm_dict = self._read_parms(ibw_wave, parm_encoding)
        self.parm_dict =  parm_dict

        data_mat = ibw_obj['wave']['wData']
        if self.parm_dict['datatype']=='image':
            data_mat = np.flip(data_mat, axis=0)
            xvec = np.linspace(0, self.parm_dict['image_width'], data_mat.shape[1])
            yvec = np.linspace(0, self.parm_dict['image_height'], data_mat.shape[0] )
            
            #make the sidpy dataset
            data_set = sid.Dataset.from_array(data_mat, name= parm_dict['channel_name'])
            data_set.data_type = 'Image'

            #Add quantity and units
            data_set.units = self.parm_dict['channel_unit']
            data_set.quantity =  self.parm_dict['parm_list'][0]

            #Add dimension info
            data_set.set_dimension(0, sid.Dimension(yvec,
                                                    name = 'x',
                                                    units=self.parm_dict['image_units'], quantity = 'x',
                                                    dimension_type='spatial'))
            data_set.set_dimension(1, sid.Dimension(xvec,
                                                    name = 'y',
                                                    units=self.parm_dict['image_units'], quantity='y',
                                                    dimension_type='spatial'))

            # append metadata
            data_set.original_metadata = self.parm_dict
        elif self.parm_dict['datatype']=='Volume CITS':
            #build spectroscopy wave
            bias_start = float(self.parm_dict['parm_list'][self.parm_dict['spec_curve_idx']].split('start =')[1].split(',')[0])
            bias_end = float(self.parm_dict['parm_list'][self.parm_dict['spec_curve_idx']].split('end =')[1].split(',')[0])
            bias_wave = np.linspace(bias_start, bias_end, self.parm_dict['spec_points_per_curve'])
            xvec = np.linspace(0, self.parm_dict['image_width'], data_mat.shape[0])
            yvec = np.linspace(0, self.parm_dict['image_height'], data_mat.shape[1])
            
            #make it sidpy
            data_set = sid.Dataset.from_array(data_mat, name= parm_dict['channel_name'])
            data_set.data_type = sid.DataType.SPECTRAL_IMAGE

            #Add quantity and units
            data_set.units = parm_dict['channel_unit']
            data_set.quantity =  parm_dict['parm_list'][0]
            
            #Add dimension info
            data_set.set_dimension(0, sid.Dimension(xvec,
                                                    name = 'x',
                                                    units=parm_dict['image_units'], quantity = 'x',
                                                    dimension_type='spatial'))
            data_set.set_dimension(1, sid.Dimension(yvec,
                                                    name = 'y',
                                                    units=parm_dict['image_units'], quantity='y',
                                                    dimension_type='spatial'))
            
            data_set.set_dimension(2, sid.Dimension(bias_wave,
                                                    name = 'Bias',
                                                    units='V', quantity='V',
                                                    dimension_type='spectral'))

            # append metadata
            data_set.original_metadata = parm_dict

        # Return the dataset
        return data_set

    @staticmethod
    def _read_parms(ibw_wave, codec='utf-8'):
        """
        Parses the parameters in the provided dictionary

        Parameters
        ----------
        ibw_wave : dictionary
            Wave entry in the dictionary obtained from loading the ibw file
        codec : str, optional
            Codec to be used to decode the bytestrings into Python strings if needed.
            Default 'utf-8'

        Returns
        -------
        parm_dict : dictionary
            Dictionary containing parameters
        """
        parm_string = ibw_wave.get('note')
        if type(parm_string) == bytes:
            try:
                parm_string = parm_string.decode(codec)
            except:
                parm_string = parm_string.decode('ISO-8859-1')  # for older AR software
        parm_string = parm_string.rstrip('\r')
        parm_list = parm_string.split('\r')
        parm_dict = dict()
        for pair_string in parm_list:
            for split_parm in [':', '=']:
                temp = pair_string.split(split_parm)
                if len(temp) == 2:
                    temp = [item.strip() for item in temp]
                    try:
                        num = float(temp[1])
                        parm_dict[temp[0]] = num
                        try:
                            if num == int(num):
                                parm_dict[temp[0]] = int(num)
                        except OverflowError:
                            pass
                    except ValueError:
                        parm_dict[temp[0]] = temp[1]

        # Grab the creation and modification times:
        other_parms = ibw_wave.get('wave_header')
        for key in ['creationDate', 'modDate', 'bname']:
            try:
                parm_dict[key] = other_parms[key]
            except KeyError:
                pass
        wh_idx =  [ind for ind in range(len(parm_list)) if 'Width' in parm_list[ind]][0]
        lines_img_idx = [ind for ind in range(len(parm_list)) if 'lines per image' in parm_list[ind]][0]
        channel_idx = [ind for ind in range(len(parm_list)) if 'Channel name' in parm_list[ind]][0]
        
        parm_dict['image_width'] = float(parm_list[wh_idx].split('=')[2][:-3])
        parm_dict['image_height'] = float(parm_list[wh_idx].split('=')[-1][:-3])
        parm_dict['image_units'] = parm_list[wh_idx].split('=')[-1][-3:][1]
        parm_dict['channel_name'] = parm_list[channel_idx].split(':')[1][:-3]
        parm_dict['channel_unit'] = parm_list[channel_idx].split(':')[1][-3:][1]

        if 'Volume CITS' in parm_dict:
            print('this is CITS') 
            spec_lines_idx = [ind for ind in range(len(parm_list)) if 'Spectroscopy points' in parm_list[ind]][0]
            sub_grid_idx = [ind for ind in range(len(parm_list)) if 'Scan Sub-Grid' in parm_list[ind]][0]
            spec_curve_idx = [ind for ind in range(len(parm_list)) if 'axis start' in parm_list[ind]][0]
            parm_dict['spec_curve_idx'] = spec_curve_idx
            parm_dict['scan_subgrid_x'] = int(re.findall(r'\d+',parm_list[sub_grid_idx].split('=')[1])[0])
            parm_dict['scan_subgrid_y'] = int(re.findall(r'\d+',parm_list[sub_grid_idx].split('=')[2])[0])
            parm_dict['spec_points_per_line'] =  int(re.findall(r'\d+', parm_list[spec_lines_idx].split('=')[1])[0])
            parm_dict['spec_lines_per_plane'] = int(re.findall(r'\d+', parm_list[spec_lines_idx].split('=')[-1])[0])
            parm_dict['spec_points_per_curve'] = int(parm_list[spec_curve_idx].split('curve =')[1].split(',')[0])
            parm_dict['lines_img_idx'] = parm_list[lines_img_idx]
            parm_dict['datatype'] = 'Volume CITS'    
            
        elif 'Image data' in parm_dict:
            print('this is an image')
            parm_dict['lines_img_idx'] = parm_list[lines_img_idx]
            parm_dict['datatype'] = 'image'    
        parm_dict['parm_list'] = parm_list
        parm_dict['other_parms'] = other_parms

        return parm_dict

    @staticmethod
    def _get_chan_labels(ibw_wave, codec='utf-8'):
        """
        Retrieves the names of the data channels and default units

        Parameters
        ----------
        ibw_wave : dictionary
            Wave entry in the dictionary obtained from loading the ibw file
        codec : str, optional
            Codec to be used to decode the bytestrings into Python strings if needed.
            Default 'utf-8'

        Returns
        -------
        labels : list of strings
            List of the names of the data channels
        default_units : list of strings
            List of units for the measurement in each channel
        """
        temp = ibw_wave.get('labels')
        labels = []
        for item in temp:
            if len(item) > 0:
                labels += item
        for item in labels:
            if item == '':
                labels.remove(item)

        default_units = list()
        for chan_ind, chan in enumerate(labels):
            # clean up channel names
            if type(chan) == bytes:
                chan = chan.decode(codec)
            if chan.lower().rfind('trace') > 0:
                labels[chan_ind] = chan[:chan.lower().rfind('trace') + 5]
            else:
                labels[chan_ind] = chan
            # Figure out (default) units
            if chan.startswith('Phase'):
                default_units.append('deg')
            elif chan.startswith('Current'):
                default_units.append('A')
            else:
                default_units.append('m')

        return labels, default_units

class IgorIBWReader(Reader):
    """
    Extracts data and metadata from Igor Binary Wave (.ibw) files containing
    images or force curves
    """
    def __init__(self, file_path, *args, **kwargs):
        if bw == None:
            raise ModuleNotFoundError('You attempted to load an Igor file, but this requires igor2.\n \
            Please Load it with pip install igor2 , restart and retry')

        super().__init__(file_path, *args, **kwargs)

    def read(self, verbose=False, parm_encoding='utf-8'):
        """
        Reads the file given in file_path into a sidpy dataset

        Parameters
        ----------
        verbose : Boolean (Optional)
            Whether or not to show  print statements for debugging
        parm_encoding : str, optional
            Codec to be used to decode the bytestrings into Python strings if
            needed. Default 'utf-8'

        Returns
        -------
        sidpy.Dataset : List of sidpy.Dataset objects.
            Multi-channel inputs are separated into individual dataset objects
        """
        file_path = self._input_file_path

        # Load the ibw file first
        ibw_obj = bw.load(file_path)
        ibw_wave = ibw_obj.get('wave')
        parm_dict = self._read_parms(ibw_wave, parm_encoding)
        chan_labels, chan_units = self._get_chan_labels(ibw_wave, parm_encoding)
        self.parm_dict =  parm_dict

        if verbose:
            print('Channels and units found:')
            print(chan_labels)
            print(chan_units)
            print('Parameters dictionary:')
            print(parm_dict)

        # Get the data to figure out if this is an image or a force curve
        images = ibw_wave.get('wData')

        datasets = {}#[] #list of sidpy datasets

        if images.shape[-1] != len(chan_labels):
            chan_labels = chan_labels[1:]  # for layer 0 null set errors in older AR software
            chan_units = chan_units[1:]

        channel_number = 0
        if images.ndim == 3:  # Image stack
            if verbose:
                print('Found image stack of size {}'.format(images.shape))

            num_rows = parm_dict['ScanLines']
            num_cols = parm_dict['ScanPoints']

            for channel in range(images.shape[-1]):
                #Convert it to sidpy dataset object
                data_set = sid.Dataset.from_array(images[:,:,channel], name=chan_labels[channel])
                data_set.data_type = 'Image'

                #Add quantity and units
                data_set.units = chan_units[channel]
                data_set.quantity = chan_labels[channel]

                #Add dimension info
                data_set.set_dimension(0, sid.Dimension(np.linspace(0, parm_dict['FastScanSize'], num_cols),
                                                        name = 'x',
                                                        units=chan_units[channel], quantity = 'x',
                                                        dimension_type='spatial'))
                data_set.set_dimension(1, sid.Dimension(np.linspace(0, parm_dict['SlowScanSize'], num_rows),
                                                        name = 'y',
                                                        units=chan_units[channel], quantity='y',
                                                        dimension_type='spatial'))

                # append metadata
                data_set.original_metadata = parm_dict
                data_set.data_type = 'image'

                #Finally, append it
                #datasets.append(data_set)
                key_channel = f"Channel_{int(channel_number):03d}"
                datasets[key_channel] = data_set
                channel_number += 1

        else:  # single force curve
            if verbose:
                print('Found force curve of size {}'.format(images.shape))

            images = np.atleast_3d(images)  # now [Z, chan, 1]

            # Find the channel that corresponds to either Z sensor or Raw:
            try:
                chan_ind = chan_labels.index('ZSnsr')
                spec_data = images[:,chan_ind].squeeze()
            except ValueError:
                try:
                    chan_ind = chan_labels.index('Raw')
                    spec_data = images[:,chan_ind,0].squeeze()
                except ValueError:
                    # We don't expect to come here. If we do, spectroscopic values remains as is
                    spec_data = np.arange(images.shape[2])

            #Go through the channels
            for channel in range(images.shape[-2]):

                # The data generated above varies linearly. Override.
                # For now, we'll shove the Z sensor data into the spectroscopic values.

                #convert to sidpy dataset
                data_set = sid.Dataset.from_array((images[:,channel,0]), name=chan_labels[channel])

                if verbose:
                    print('Channel {} and spec_data is {}'.format(channel, spec_data))

                #Set units, quantity
                data_set.units = chan_units[channel]
                data_set.quantity = chan_labels[channel]

                data_set.set_dimension(0, sid.Dimension(spec_data, name = chan_labels[channel],
                                                        units=chan_units[channel], quantity=chan_labels[channel],
                                                        dimension_type='spectral'))

                #append metadata
                data_set.data_type = 'SPECTRUM'
                data_set.original_metadata = parm_dict

                #Add dataset to list
                #datasets.append(data_set)
                key_channel = f"Channel_{int(channel_number):03d}"
                datasets[key_channel] = data_set
                channel_number += 1

        # Return the dataset
        return datasets

    @staticmethod
    def _read_parms(ibw_wave, codec='utf-8'):
        """
        Parses the parameters in the provided dictionary

        Parameters
        ----------
        ibw_wave : dictionary
            Wave entry in the dictionary obtained from loading the ibw file
        codec : str, optional
            Codec to be used to decode the bytestrings into Python strings if needed.
            Default 'utf-8'

        Returns
        -------
        parm_dict : dictionary
            Dictionary containing parameters
        """
        parm_string = ibw_wave.get('note')
        if type(parm_string) == bytes:
            try:
                parm_string = parm_string.decode(codec)
            except:
                parm_string = parm_string.decode('ISO-8859-1')  # for older AR software
        parm_string = parm_string.rstrip('\r')
        parm_list = parm_string.split('\r')
        parm_dict = dict()
        for pair_string in parm_list:
            for split_parm in [':', '=']:
                temp = pair_string.split(split_parm)
                if len(temp) == 2:
                    temp = [item.strip() for item in temp]
                    try:
                        num = float(temp[1])
                        parm_dict[temp[0]] = num
                        try:
                            if num == int(num):
                                parm_dict[temp[0]] = int(num)
                        except OverflowError:
                            pass
                    except ValueError:
                        parm_dict[temp[0]] = temp[1]

        # Grab the creation and modification times:
        other_parms = ibw_wave.get('wave_header')
        for key in ['creationDate', 'modDate', 'bname']:
            try:
                parm_dict[key] = other_parms[key]
            except KeyError:
                pass
        return parm_dict

    @staticmethod
    def _get_chan_labels(ibw_wave, codec='utf-8'):
        """
        Retrieves the names of the data channels and default units

        Parameters
        ----------
        ibw_wave : dictionary
            Wave entry in the dictionary obtained from loading the ibw file
        codec : str, optional
            Codec to be used to decode the bytestrings into Python strings if needed.
            Default 'utf-8'

        Returns
        -------
        labels : list of strings
            List of the names of the data channels
        default_units : list of strings
            List of units for the measurement in each channel
        """
        temp = ibw_wave.get('labels')
        labels = []
        for item in temp:
            if len(item) > 0:
                labels += item
        for item in labels:
            if item == '':
                labels.remove(item)

        default_units = list()
        for chan_ind, chan in enumerate(labels):
            # clean up channel names
            if type(chan) == bytes:
                chan = chan.decode(codec)
            if chan.lower().rfind('trace') > 0:
                labels[chan_ind] = chan[:chan.lower().rfind('trace') + 5]
            else:
                labels[chan_ind] = chan
            # Figure out (default) units
            if chan.startswith('Phase'):
                default_units.append('deg')
            elif chan.startswith('Current'):
                default_units.append('A')
            else:
                default_units.append('m')

        return labels, default_units

    def can_read(self):
        """
        Tests whether or not the provided file has a .ibw extension
        Returns
        -------

        """

        return super(IgorIBWReader, self).can_read(extension='ibw')





from modulefinder import Module
import numpy as np  # For array operations
import sidpy as sid
from sidpy.sid import Reader
from os import path
import sys

if sys.version_info.major == 3:
    unicode = str

try:
    import gwyfile
except ModuleNotFoundError:
    print("You don't have gwyfile installed. \
    If you wish to open .gwy files, you will need to \
     install it (pip install gwyfile) before attempting.")
    gwyfile = None

class GwyddionReader(Reader):
    """
    Extracts data and metadata from Igor Binary Wave (.ibw) files containing
    images or force curves
    """
    def __init__(self, file_path, *args, **kwargs):
        if gwyfile == None:
            raise ModuleNotFoundError('You attempted to load a Gwyddion file, but this requires gwyfile.\n \
    Please Load it with pip install gwyfile , restart and retry')
        super().__init__(file_path, *args, **kwargs)

    def read(self, verbose=False, parm_encoding='utf-8'):
        """
        Reads the file given in file_path into a sidpy dataset

        Parameters
        ----------
        verbose : Boolean (Optional)
            Whether or not to show  print statements for debugging
     
        Returns
        -------
        sidpy.Dataset : List of sidpy.Dataset objects.
            Multi-channel inputs are separated into individual dataset objects
        """
        file_path = self._input_file_path

        if not isinstance(file_path, (str, unicode)):
            raise TypeError('file_path should be a string!')
        if not (file_path.endswith('.gsf') or file_path.endswith('.gwy')):
            # TODO: Gwyddion is weird, it doesn't append the file extension some times.
            # In theory, you could identify the kind of file by looking at the header (line 38 in gsf_read()).
            # Ideally the header check should be used instead of the extension check
            raise ValueError('file_path must have a .gsf or .gwy extension!')
        
        if file_path.endswith('.gsf'):
            datasets = self.gsf_read()
        elif file_path.endswith('.gwy'):
            datasets = self.gwy_read()
        
        # Return the dataset
        return datasets

    def gsf_read(self):
        """
        Read a Gwyddion Simple Field 1.0 file format
        http://gwyddion.net/documentation/user-guide-en/gsf.html
        
        Parameters
        ----------
        file_name : string
            path to the file
        Returns
        -------
        metadata : dict)
            Additional metadata to be included in the file
        data : numpy.ndarray
            An arbitrarily sized 2D array of arbitrary numeric type
        """
       
        gsf_file = open(self._input_file_path, 'rb')
        
        metadata = {}
        
        # check if header is OK
        if not(gsf_file.readline().decode('UTF-8') == 'Gwyddion Simple Field 1.0\n'):
            gsf_file.close()
            raise ValueError('File has wrong header')
            
        term = b'00'
        # read metadata header
        while term != b'\x00':
            line_string = gsf_file.readline().decode('UTF-8')
            metadata[line_string.rpartition(' = ')[0]] = line_string.rpartition('=')[2]
            term = gsf_file.read(1)
            gsf_file.seek(-1, 1)
        
        gsf_file.read(4 - gsf_file.tell() % 4)
        
        # fix known metadata types from .gsf file specs
        # first the mandatory ones...
        metadata['XRes'] = int(metadata['XRes'])
        metadata['YRes'] = int(metadata['YRes'])
        
        # now check for the optional ones
        if 'XReal' in metadata:
            metadata['XReal'] = float(metadata['XReal'])
        
        if 'YReal' in metadata:
            metadata['YReal'] = float(metadata['YReal'])
                    
        if 'XOffset' in metadata:
            metadata['XOffset'] = float(metadata['XOffset'])
        
        if 'YOffset' in metadata:
            metadata['YOffset'] = float(metadata['YOffset'])
        
        datasets = []

        data = np.frombuffer(gsf_file.read(), dtype='float32').reshape(metadata['YRes'], metadata['XRes'])
        
        gsf_file.close()
        num_cols, num_rows = data.shape
        data_set = sid.Dataset.from_array(data, title=metadata['Title'])
        data_set.data_type = 'Image'

        #Add quantity and units
        
        try:
            data_set.units = metadata['ZUnits']
        except Exception as e:
            # If an error occurs, fall back to using "a.u"
            data_set.units = "a.u"
        data_set.quantity = metadata['Title']
        
        try:
            units = metadata['XYUnits'],
        except Exception as e:
            # If an error occurs, fall back to using "a.u"
            units= "a.u"

        #Add dimension info
        data_set.set_dimension(0, sid.Dimension(np.linspace(0, metadata['XReal'], num_cols),
                                                name = 'x',
                                                units= units,
                                                quantity = 'x',
                                                dimension_type='spatial'))
        data_set.set_dimension(1, sid.Dimension(np.linspace(0, metadata['YReal'], num_rows),
                                                name = 'y',
                                                units = units,
                                                quantity = 'y',
                                                dimension_type='spatial'))

        # append metadata
        data_set.original_metadata = metadata
        data_set.data_type = 'image'

        #Finally, append it
        datasets.append(data_set)
        
        return datasets


    def gwy_read(self):
        """
        Parameters
        ----------
        file_path
        meas_grp
        For more information on the .gwy file format visit the link below -
        http://gwyddion.net/documentation/user-guide-en/gwyfile-format.html
        """

        # Need to build a set of channels to test against and a function-level variable to write to
        channels = []

        # Read the data in from the specified file
        gwy_data = gwyfile.load(self._input_file_path)

        for obj in gwy_data:
            gwy_key = obj.split('/')
            try:
                # if the second index of the gwy_key can be cast into an int then
                # it needs to be processed either as an image or a graph
                
                int(gwy_key[1])
                
                if gwy_key[2] == 'graph':
                    # graph processing
                   
                    channels = self._translate_graph(gwy_data,
                                                        obj)
                elif obj.endswith('data'):
                   
                    channels.append(self._translate_image_stack( gwy_data,
                                                            obj))
                else:
                    continue
            except ValueError:
                # if the second index of the gwy_key cannot be cast into an int
                # then it needs to be processed wither as a spectra, volume or xyz
                
                if gwy_key[1] == 'sps':
                
                    channels = self._translate_spectra(gwy_data,
                                                        obj)
                elif gwy_key[1] == 'brick':
                 
                    channels = self._translate_volume(gwy_data,
                                                        obj)
                elif gwy_key[1] == 'xyz':
                 
                    channels = self._translate_xyz(gwy_data,obj)
         
        return channels


    def _translate_image_stack(self, gwy_data, obj):
        """
        Use this function to write data corresponding to a stack of scan images (most common)
        Returns
        -------
        """        
    
        if obj.endswith('data'):
            data = gwy_data[obj].data
            title = gwy_data[obj+'/title']
            keys = gwy_data[obj].keys()
            meta_data = {}
            for key in keys:
                if 'data' not in key:
                    meta_data[key] = gwy_data[obj][key] 

            x_range = gwy_data[obj].get('xreal')
            x_vals = np.linspace(0, x_range, gwy_data[obj]['xres'])
          
            y_range = gwy_data[obj].get('yreal')
            y_vals = np.linspace(0, y_range, gwy_data[obj]['yres'])

            data_set = sid.Dataset.from_array(data, title=title)
            data_set.data_type = 'Image'

            #Add quantity and units
            if 'si_unit_z' in gwy_data[obj].keys(): 
                data_set.units = gwy_data[obj]['si_unit_z']['unitstr']
            else:
                data_set.units = 'a.u.'

            data_set.quantity = title 

            if 'si_unit_xy' in gwy_data[obj].keys():
                x_unit = y_unit = gwy_data[obj]['si_unit_xy']['unitstr']
            else:
                x_unit=y_unit='a.u.'

            #Add dimension info
            dataset_is_transposed = False
            if data_set.shape[0] != len(x_vals):
                dataset_is_transposed = True

            if not dataset_is_transposed:

                data_set.set_dimension(0, sid.Dimension(x_vals,
                                                        name = 'x',
                                                        units=x_unit, quantity = 'x',
                                                        dimension_type='spatial'))
                data_set.set_dimension(1, sid.Dimension(y_vals,
                                                        name = 'y',
                                                        units=y_unit, quantity = 'y',
                                                        dimension_type='spatial'))
            else:
                data_set.set_dimension(0, sid.Dimension(y_vals,
                                                        name = 'y',
                                                        units=y_unit, quantity = 'y',
                                                        dimension_type='spatial'))
                
                data_set.set_dimension(1, sid.Dimension(x_vals,
                                                        name = 'x',
                                                        units=x_unit, quantity = 'x',
                                                        dimension_type='spatial'))

            # append metadata
            data_set.original_metadata = meta_data
            data_set.data_type = 'image'
       
        return data_set


    def _translate_spectra(self, gwy_data, obj):
        """
        Use this to translate simple 1D data like force curves
        Returns
        -------
        """
        keys = gwy_data[obj].keys()
        meta_data = {}
        for key in keys:
            if 'data' not in key:
                meta_data[key] = gwy_data[obj][key] 

        title = obj['title']
        unitstr = obj['unitstr']
        coords = obj['coords']
        res = obj['data']['res']
        real = obj['data']['real']
        offset = obj['data']['off']
        x_units = obj['data']['si_unit_x']['unitstr']
        y_units = obj['data']['si_unit_y']['unitstr']
        data = obj['data']['data']
        indices = obj['selected']
        x_vals = np.linspace(offset, real, res)

        data_set = sid.Dataset.from_array(data, title=title)
        data_set.data_type = 'line_plot'

        #Add quantity and units
        data_set.units = y_units
        data_set.quantity = title 
    
        #Add dimension info
        data_set.set_dimension(0, sid.Dimension(x_vals,
                                                name = 'x',
                                                units = x_units, quantity = 'x',
                                                dimension_type='spectral'))

        # append metadata
        data_set.original_metadata = meta_data

        return data_set


    #TODO: This stuff is not yet supported. Need to work on it...

    def _translate_graph(self,  gwy_data, obj):
        """
        Use this to translate graphs
        Returns
        """
        return 


    def _translate_volume(self, gwy_data, obj):
        return 


    def _translate_xyz(self, gwy_data, obj):
        return 



"""
Author: Pranav Sudersan
Date: February 2025

This module provides classes and functions for reading and processing WSxM/Nanotec AFM controller files.
It includes SID readers for 1D, 2D, and 3D data formats as three separate classes:

Classes:
--------
WSxM2DReader(Reader)
    A reader class for extracting 2D WSxM data and metadata (e.g., topography, phase, adhesion, etc.)
    and converting them into a dictionary of SID Dataset objects for all channels and directions
    of the measurement.

WSxM1DReader(Reader)
    A reader class for extracting 1D WSxM data and metadata (e.g., force spectroscopy, tune, etc.)
    and converting them into a dictionary of SID Dataset objects for all channels and directions
    of the measurement.

WSxM3DReader(Reader)
    A reader class for extracting 3D WSxM data and metadata (e.g., force volume, video, etc.)
    and converting them into a dictionary of SID Dataset objects for all channels and directions
    of the measurement.

WSxMFuncs
    A class containing optional static methods to directly read WSxM files, including reading headers,
    image data, and spectroscopy curves without the SID format. These methods are used by the above
    WSxM reader classes but can also be used independently for reading WSxM files, and they offer
    additional input parameters.

Usage:
------
To use the readers, instantiate the appropriate reader class with the file path and call the
read method to extract the data and metadata.

Example:
--------
    data_file_path = '/path/to/wsxm/file'
    reader = WSxM2DReader(data_file_path)
    data = reader.read()
    for channel, dataset in data.items():
        print(channel, dataset.quantity, dataset.direction)
        dataset.plot()
"""

import struct
import re
import sys
import numpy as np
import sidpy as sid
from sidpy.sid import Reader
from pathlib import Path
from matplotlib import pyplot as plt


# Read two dimensional AFM image data (e.g. Topography, Phase etc.)
class WSxM2DReader(Reader):
    def __init__(self, file_path, *args, **kwargs):
        """
        A reader class for extracting 2 dimensional WSxM data and metadata (e.g. topography, 
        phase, adhesion etc) and converting them into a dictionary of SID Dataset objects 
        for all channels and directions of the measurement.

        Parameters
        ----------
        file_path : str
            Path to the input file to be read.
        *args : tuple, optional
            Additional arguments to pass to the parent sidpy Reader class.
        **kwargs : dict, optional
            Additional keyword arguments to pass to the parent sidpy Reader class.

        Methods
        -------
        read()
            Extracts the data and metadata from the input 2D data file and returns it as
            a dictionary of sidpy.Dataset objects.
        """
        super().__init__(file_path, *args, **kwargs)


    def read(self):
        """
        Reads and processes 2D WSxM files.
        Returns:
            dict: A dictionary where keys are channel identifiers ("Channel_xxx") and values are 
                    sidpy.Dataset objects containing the extracted data and metadata.
        """

        filepath = Path(self._input_file_path)
        filepath_all = WSxMFuncs._wsxm_get_common_files(filepath) #get all files of measurement, with common base name
        datasets = {}
        channel_number = 0 #channel number
        for path_i in filepath_all:
            path_ext = path_i.suffix  #file extension
            if path_ext != '.gsi': #ignore *.gsi files sharing same name               
                file = open(f'{path_i}','rb')
                header_dict, pos = WSxMFuncs._wsxm_readheader(file)
                header_dict['File path'] = path_i #file path included to header
                chan_label = header_dict['Acquisition channel [General Info]']
                data_dict_chan, pos = WSxMFuncs._wsxm_readimg(file, header_dict, pos)
                file.close()
                if 'X scanning direction [General Info]' in header_dict.keys():
                    x_dir = header_dict['X scanning direction [General Info]']
                else:
                    x_dir = None

                data_set = sid.Dataset.from_array(np.flip(data_dict_chan['data']['Z'].T),
                                                  title=chan_label)
                
                #Add quantity and units
                data_set.units = data_dict_chan['units']['Z']
                data_set.quantity = chan_label
                data_set.direction = x_dir #image direction information
                data_set.data_type = 'image'                

                #Add dimension info
                data_set.set_dimension(0, sid.Dimension(data_dict_chan['data']['X'],
                                                        name = 'x',
                                                        units=data_dict_chan['units']['X'], 
                                                        quantity = 'x',
                                                        dimension_type='spatial'))
                data_set.set_dimension(1, sid.Dimension(data_dict_chan['data']['Y'],
                                                        name = 'y',
                                                        units=data_dict_chan['units']['Y'], 
                                                        quantity='y',
                                                        dimension_type='spatial')) 
                #Writing the metadata
                data_set.metadata = data_dict_chan['header'].copy()

                #Add dataset to dictionary
                key_channel = f"Channel_{int(channel_number):03d}"
                datasets[key_channel] = data_set
                channel_number += 1

        return datasets

# Read one dimensional AFM data (e.g. force-distance curves)        
class WSxM1DReader(Reader):
    def __init__(self, file_path, *args, **kwargs):
        """
        A reader class for extracting 1 dimensional WSxM data and metadata (e.g. force 
        spectroscopy, tune etc) and converting them into a dictionary of SID Dataset objects 
        for all channels and directions of the measurement.

        Parameters
        ----------
        file_path : str
            Path to the input file to be read.
        *args : tuple, optional
            Additional arguments to pass to the parent sidpy Reader class.
        **kwargs : dict, optional
            Additional keyword arguments to pass to the parent sidpy Reader class.

        Methods
        -------
        read()
            Extracts the data and metadata from the input 1D data file and returns it as
            a dictionary of sidpy.Dataset objects.
        """    
        super().__init__(file_path, *args, **kwargs)

    def read(self):
        """
        Reads and processes 1D WSxM files.
        Returns:
            dict: A dictionary where keys are channel identifiers ("Channel_xxx") and values are 
                    sidpy.Dataset objects containing the extracted data and metadata.
        """
        filepath = Path(self._input_file_path)
        filepath_all = WSxMFuncs._wsxm_get_common_files(filepath) #get all files of measurement, with common base name

        data_dict_chan = {}
        data_dict_stp = {} #dictionary for *.stp files to combine approach and retract data from separate files
        datasets = {}
        channel_number = 0 #channel number
        for path_i in filepath_all:
            path_ext = path_i.suffix                
            if path_ext == '.curves': # read *.curves spectroscopy files
                temp_dict, chan_label = WSxMFuncs._wsxm_readcurves(path_i)
                if chan_label not in data_dict_chan.keys():
                    data_dict_chan[chan_label] = temp_dict[chan_label].copy()
                else:
                    for curv_ind_i in temp_dict[chan_label]['curves'].keys(): #replace with *.curves data even if it already exists (more robust)
                        data_dict_chan[chan_label]['curves'][curv_ind_i] = temp_dict[chan_label]['curves'][curv_ind_i].copy()
            elif path_ext == '.stp': # read *.stp spectroscopy files
                temp_dict, chan_label = WSxMFuncs._wsxm_readstp(path_i, data_dict_stp)
                if chan_label not in data_dict_chan.keys(): #ignore data if *.curves already found
                    data_dict_chan[chan_label] = temp_dict[chan_label].copy()
                else:
                    for curv_ind_i in temp_dict[chan_label]['curves'].keys():
                        if curv_ind_i not in data_dict_chan[chan_label]['curves'].keys():
                            data_dict_chan[chan_label]['curves'][curv_ind_i] = temp_dict[chan_label]['curves'][curv_ind_i].copy()
            elif path_ext == '.cur': # read *.cur spectroscopy files
                temp_dict, chan_label = WSxMFuncs._wsxm_readcur(path_i)
                if chan_label not in data_dict_chan.keys(): #ignore data if *.curves already found
                    data_dict_chan[chan_label] = temp_dict[chan_label].copy()
                else:
                    for curv_ind_i in temp_dict[chan_label]['curves'].keys():
                        if curv_ind_i not in data_dict_chan[chan_label]['curves'].keys():
                            data_dict_chan[chan_label]['curves'][curv_ind_i] = temp_dict[chan_label]['curves'][curv_ind_i].copy()        

        #convert data dictionary to SID dataset
        for chan_i, chandata_i in data_dict_chan.items():
            for curv_i, curvdata_i in chandata_i['curves'].items():
                curve_list = []
                curve_dir_list = []
                for curv_dir_i, curvdata_dir_i in curvdata_i['data'].items():
                    curve_list.append(curvdata_dir_i['y'])
                    curve_dir_list.append(curv_dir_i)
                    x_data_i = curvdata_dir_i['x']
                curve_matrix = np.vstack(curve_list).T

                data_set = sid.Dataset.from_array(curve_matrix, title=f'{chan_i} ({curv_i})')
                
                #Add quantity and units
                data_set.units = curvdata_i['units']['y']
                data_set.quantity = chan_i
                data_set.direction = curve_dir_list #image direction information
                data_set.data_type = 'spectrum'                

                #Add dimension info
                data_set.set_dimension(0, sid.Dimension(x_data_i,
                                                        name = 'x',
                                                        units=curvdata_i['units']['x'],
                                                        quantity = 'x',
                                                        dimension_type='spatial'))
                
                #Writing the metadata
                data_set.metadata = curvdata_i['header'].copy()

                #Add dataset to dictionary
                key_channel = f"Channel_{int(channel_number):03d}"
                datasets[key_channel] = data_set
                channel_number += 1

        return datasets


# Read three dimensional AFM image data (e.g. Force volume, video etc)
class WSxM3DReader(Reader):
    def __init__(self, file_path, *args, **kwargs):
        """
        A reader class for extracting 3 dimensional WSxM data and metadata (e.g. Force 
        volume, video etc) and converting them into a dictionary of SID Dataset objects 
        for all channels and directions of the measurement.

        Parameters
        ----------
        file_path : str
            Path to the input file to be read.
        *args : tuple, optional
            Additional arguments to pass to the parent sidpy Reader class.
        **kwargs : dict, optional
            Additional keyword arguments to pass to the parent sidpy Reader class.

        Methods
        -------
        read()
            Extracts the data and metadata from the input 3D data file and returns it as
            a dictionary of sidpy.Dataset objects.
        """
        super().__init__(file_path, *args, **kwargs)


    def read(self):
        """
        Reads and processes 3D WSxM files.
        Returns:
            dict: A dictionary where keys are channel identifiers ("Channel_xxx") and values are 
                    sidpy.Dataset objects containing the extracted data and metadata.
        """

        filepath = Path(self._input_file_path)
        filepath_all = WSxMFuncs._wsxm_get_common_files(filepath) #get all files of measurement, with common base name

        datasets = {}
        channel_number = 0 #channel number
        for path_i in filepath_all:
            path_ext = path_i.suffix
            if path_ext == '.gsi': #force volume data from *.gsi files
                data_dict_chan, chan_label, topo_data = WSxMFuncs._wsxm_readforcevol(path_i)
                zz_data_type = sid.DataType.SPECTRAL_IMAGE 
                z_dimension_type = 'spectral'
            elif path_ext in ['.MOV', '.mpp']: #movie data files
                data_dict_chan, chan_label = WSxMFuncs._wsxm_readmovie(path_i)
                zz_data_type = sid.DataType.IMAGE_STACK  
                z_dimension_type = 'frame'
            else:
                assert("Invalid file type. Please choose *.gsi/*.MOV/*.mpp file")

            #image rotated to correct orientation of measurement in final dataset
            data_set = sid.Dataset.from_array(np.flip(np.rot90(data_dict_chan['data']['ZZ'], k=1, axes=(2,1)), axis=1),
                                                title=chan_label)
            
            #Add quantity and units
            data_set.units = data_dict_chan['units']['ZZ']
            data_set.quantity = chan_label
            data_set.data_type = zz_data_type            

            #Add dimension info
            data_set.set_dimension(0, sid.Dimension(data_dict_chan['data']['Z'],
                                                    name = 'z',
                                                    units=data_dict_chan['units']['Z'], 
                                                    quantity = 'z',
                                                    dimension_type=z_dimension_type))
            data_set.set_dimension(1, sid.Dimension(data_dict_chan['data']['X'],
                                                    name = 'x',
                                                    units=data_dict_chan['units']['X'], 
                                                    quantity='x',
                                                    dimension_type='spatial'))
            data_set.set_dimension(2, sid.Dimension(data_dict_chan['data']['Y'],
                                                    name = 'y',
                                                    units=data_dict_chan['units']['Y'], 
                                                    quantity='y',
                                                    dimension_type='spatial')) 
            
            #Writing the metadata            
            header_spectroscopy = {k: v for k, v in data_dict_chan['header'].items() if "[Spectroscopy images ramp value list]" in k or "[maxmins list]" in k}
            header_general = {k: v for k, v in data_dict_chan['header'].items() if k not in header_spectroscopy.keys()}
            data_set.metadata = header_general.copy()
            data_set.metadata['Spectroscopy metadata'] = header_spectroscopy
            if zz_data_type == sid.DataType.SPECTRAL_IMAGE: 
                data_set.direction = data_dict_chan['header']['Spectroscopy type [General Info]'] #image direction information
                #create topography dataset
                data_set_topo = sid.Dataset.from_array(np.flip(topo_data.T), title='Topography')
                data_set_topo.set_dimension(0, sid.Dimension(data_dict_chan['data']['X'],
                                                    name = 'x',
                                                    units=data_dict_chan['units']['X'], 
                                                    quantity='x',
                                                    dimension_type='spatial'))
                data_set_topo.set_dimension(1, sid.Dimension(data_dict_chan['data']['Y'],
                                                        name = 'y',
                                                        units=data_dict_chan['units']['Y'], 
                                                        quantity='y',
                                                        dimension_type='spatial'))
                data_set_topo.units = data_dict_chan['units']['Z']
                data_set_topo.quantity = 'Topography'
                data_set_topo.direction = data_dict_chan['header']['Spectroscopy type [General Info]'].split(' ')[1] #image direction information
                data_set_topo.data_type = 'image'

                data_set.metadata['Topography'] = data_set_topo #topography data added to metadata
                
            elif zz_data_type == sid.DataType.IMAGE_STACK:
                data_set.direction = 'None'


            #Add dataset to dictionary
            key_channel = f"Channel_{int(channel_number):03d}"
            datasets[key_channel] = data_set
            channel_number += 1

        return datasets

# collection of general functions used in WSxM readers
class WSxMFuncs():
    """
    A class containing static methods for handling WSxM files, including reading headers,
    image data, and spectroscopy curves. These methods are used by the WSxM reader classes
    above but can also be used by itself for reading WSxM files, offering additional input 
    parameters.
    Methods
    -------
    _wsxm_get_common_files(filepath, ext=None)
    _wsxm_readheader(file, pos=0, inibyte=100)
    _wsxm_readimg(file, header_dict, pos)
    _wsxm_readcurves(filepath)
    """
    # binary data type definitions used in header file
    DATA_TYPES = {
        'short':(2,'h'),'short-data':(2,'h'), 'unsignedshort':(2,'H'),
        'integer-data':(4,'i'), 'signedinteger':(4,'i'),
        'float-data':(4,'f'), 'double':(8,'d')
                }
    
    #rename spectroscopy line to standard names: approach and retract
    SPECT_DICT = {
        'Forward': 'approach', 'Backward': 'retract',
        'b': 'retract', 'f': 'approach'
                } 

    
    @staticmethod
    def _wsxm_get_common_files(filepath, ext=None):
        """
        Retrieve a list of files corresponding to all the channels of the 
        chosen measurement file (including "Forward" and "Backward" data), if they exist.
        
        This function searches for files in the same directory as the provided
        `filepath` that share a common base name, identified by a 4-digit number
        in the filename (standard WSxM naming convention). The provided `filepath` 
        is always included as the first item in the returned list. Optionally, the 
        search can be filtered by file extension.
        
        Args:
            filepath (Path): The path to the reference file.
            ext (str, optional): The file extension to filter by (e.g. ".top", ".ch1", ".gsi" etc.). 
                     If None, all files with the common base name are included.
        
        Returns:
            list[Path]: A list of Path objects representing the files for all channels of the measurement, 
                with the provided `filepath` as the first item.
        
        Note:
            - If the data files were renamed later, this function may not work as expected. 
              It is recommended to avoid renaming the part of the file name before the 4-digit number.
        """

        path_dir = filepath.parent
        filename = filepath.name
        match = re.search(r'\_\d{4}', filename) #regex to find 4 digit number in filename
        if match == None: #return same file if no matches
            return [filepath]
        else:
            filename_com = filename[:match.start()+5]

        files = []
        for path_i in path_dir.iterdir():
            path_ext_i = path_i.suffix
            if ext != None and path_ext_i != ext: #if ext given, skip files dont match the extension
                continue
            if path_i.is_file() and path_i.name.startswith(filename_com):
                files.append(path_i) 
        files.remove(filepath) #make sure filepath is the first item in the list
        files.insert(0, filepath)
        return files


    #read WSxM header data
    @staticmethod
    def _wsxm_readheader(file, pos=0, inibyte=100):
        """
        Reads the header of a WSxM file and returns it as a dictionary.
        Parameters:
        file (file object): The file object to read from.
        pos (int, optional): The position in the file (in bytes) to start reading from. Defaults to 0.
        inibyte (int, optional): The initial number of bytes to read to find 'Image header size'. Defaults to 100.
        Returns:
            tuple: A tuple containing:
                - header_dict (dict): A dictionary containing the header information.
                - pos_new (int): The new position (in bytes) in the file after reading the header.
        """

        header_dict = {}
        title_list = []
        # Find header size
        file.seek(pos, 0)
        data = file.read(inibyte)
        for ln in data.splitlines():
            hd_lst = ln.decode('latin-1', errors='ignore').split(':')
            if len(hd_lst) == 2:
                if hd_lst[0] == 'Image header size':
                    header_size = int(hd_lst[1])
                    break
        # read header data
        file.seek(pos, 0)
        data = file.read(header_size)
        for ln in data.splitlines():
            hd_lst = ln.decode('latin-1', errors='ignore').split(':')
            if len(hd_lst) == 2:
                header_name = f"{hd_lst[0].strip()} {title_list[-1]}".strip()
                header_dict[header_name] = hd_lst[1].strip()
            elif len(hd_lst) == 1 and hd_lst[0] != '': #collect section tiles in header file
                title_list.append(hd_lst[0])
        
        pos_new = pos + header_size #bytes read so far
        return header_dict, pos_new

    #read WSxM binary image data
    @staticmethod
    def _wsxm_readimg(file, header_dict, pos):
        """
        Reads an image from a WSxM file and returns the image data along with updated position.
        Parameters:
        file (file object): The file object to read the image data from.
        header_dict (dict): Dictionary containing the header information of the WSxM file.
        pos (int): The current position in the file (in bytes) from where to start reading the image data.
        Returns:
            tuple: A tuple containing:
                - data_dict_chan (dict): A dictionary with the image data and header information.
                    - 'data' (dict): Contains the image data arrays.
                        - 'Z' (numpy.ndarray): The 2D image data (image data) calibrated and reshaped.
                        - 'X' (numpy.ndarray): The 1D X-axis data.
                        - 'Y' (numpy.ndarray): The 1D Y-axis data.
                    - 'header' (dict): A copy of the header dictionary.
                    - 'unit' (dict): Units for 'Z', 'X' and 'Y' data
                - pos (int): The updated position in the file after reading the image data.
        """

        data_format = header_dict['Image Data Type [General Info]']
        chan_label = header_dict['Acquisition channel [General Info]']
        x_num = int(header_dict['Number of rows [General Info]'])
        y_num = int(header_dict['Number of columns [General Info]'])
        x_len = float(header_dict['X Amplitude [Control]'].split(' ')[0])
        y_len = float(header_dict['Y Amplitude [Control]'].split(' ')[0])
        z_len = float(header_dict['Z Amplitude [General Info]'].split(' ')[0])
        z_min = float(header_dict['Minimum [Miscellaneous]'])
        z_max = float(header_dict['Maximum [Miscellaneous]'])
        if 'X starting offset [General Info]' in header_dict.keys(): #for "3D mode" images
            x_offset = float(header_dict['X starting offset [General Info]'].split(' ')[0])
            y_offset = float(header_dict['Y starting offset [General Info]'].split(' ')[0])
        else:
            x_offset = 0
            y_offset = 0

        x_data = np.linspace(x_offset, x_len+x_offset, x_num, endpoint=True)
        y_data = np.linspace(y_len+y_offset, y_offset, y_num, endpoint=True)
        
        #read binary image data
        point_length, type_code  = WSxMFuncs.DATA_TYPES[data_format]
        file.seek(pos, 0)
        data_len = x_num*y_num*point_length
        bin_data = file.read(data_len)
        ch_array = np.array(list(struct.iter_unpack(f'{type_code}', bin_data))).flatten()

        if z_len == 0: #for zero data
            z_calib = 1
        else:
            z_calib = z_len/(z_max-z_min)
        
        z_unit = header_dict['Z Amplitude [General Info]'].split(' ')[-1]

        #the following fixes a bug in the data format for amplitude channel, ensures that data is read in volts, not nanometers (which is fake)
        if chan_label == 'Amplitude' and header_dict['Z Amplitude [General Info]'].split(' ')[1] != 'V':
            chan_factor = float(header_dict['Conversion Factor 00 [General Info]'].split(' ')[0])
            z_calib = z_calib/chan_factor
            z_unit = 'V'
        
        #img data dictionary
        data_dict_chan = {'data': {'Z': (z_calib*ch_array.reshape(x_num, y_num)),
                                'X': x_data,
                                'Y': y_data},
                        'header': header_dict.copy(),
                        'units': {'Z': z_unit,
                                  'X': header_dict['X Amplitude [Control]'].split(' ')[-1],
                                  'Y': header_dict['Y Amplitude [Control]'].split(' ')[-1]
                                  }
                        }
        
        pos += data_len #bytes read so far
        
        return data_dict_chan, pos

    # read *.curves file with image and f-d curves
    @staticmethod
    def _wsxm_readcurves(filepath):
        """
        Reads WSxM spectroscopy curves (*.curves format) from a given file.
        Args:
            filepath (str): The path to the WSxM file to be read.
        Returns:
            tuple: A tuple containing:
            - data_dict (dict): A dictionary containing the parsed data from the file.
            - y_label (str): The channel name of the Y-axis data.
        The structure of the returned data_dict is as follows:
        {
            'Y axis channel label': {
            'image': {
                'data': {
                'Z': 2D array of embedded topography data,
                'X': 1D array of X-axis data,
                'Y': 1D array of Y-axis data
                },
                'header': header_dict
            },
            'curves': {
                curve_index: {
                'header': header_dict,
                'units': {'x': x_unit, 'y': y_unit},
                'data': {
                    'approach': {'x': x_data, 'y': y_data},
                    'retract': {'x': x_data, 'y': y_data}
                }
                }
            }
            }
        }
        """

        data_dict = {}
        file = open(f'{filepath}','rb')
        header_dict_top, pos = WSxMFuncs._wsxm_readheader(file)
        data_dict_chan, pos = WSxMFuncs._wsxm_readimg(file, header_dict_top, pos) 
        
        data_format = header_dict_top['Image Data Type [General Info]']
        point_length, type_code  = WSxMFuncs.DATA_TYPES[data_format]
        data_dict_curv = {}
        
        while True:
            header_dict, pos = WSxMFuncs._wsxm_readheader(file, pos=pos)
            header_dict['File path'] = filepath #file path included to header 
            line_pts = int(header_dict['Number of points [General Info]'])
            line_num = int(header_dict['Number of lines [General Info]'])
            y_label = header_dict['Y axis text [General Info]'].split('[')[0].strip()
            x_label = header_dict['X axis text [General Info]'].split('[')[0].strip()
            curv_ind = int(header_dict['Index of this Curve [Control]'])
            curv_num = int(header_dict['Number of Curves in this serie [Control]'])
            chan_fact = float(header_dict['Conversion Factor 00 [General Info]'].split(' ')[0])
            chan_inv = header_dict['Channel is inverted [General Info]']
            if chan_inv == 'Yes':
                chan_fact = -chan_fact
            chan_offs = float(header_dict['Conversion Offset 00 [General Info]'].split(' ')[0])

            header_dict['Spectroscopy channel'] = y_label #Insert channel name information into dictionary
            
            line_order = ['approach', 'retract']
            if header_dict['First Forward [Miscellaneous]'] == 'No': #CHECK THIS
                line_order = ['retract', 'approach']

            data_len = line_pts*line_num*2*point_length
            file.seek(pos, 0)
            
            if line_pts == 0: #skip if no data for curve exists (bug in file format)
                continue
            
            bin_data = file.read(data_len)
            ch_array = np.array(list(struct.iter_unpack(f'{type_code}', bin_data))).flatten()
            x_data = np.split(ch_array[::2], line_num)
            y_data = np.split(ch_array[1::2], line_num)

            for j in range(int(line_num/len(line_order))):
                k = len(line_order) * j
                curv_ind_j = f'{curv_ind}{chr(ord("a")+j)}' if line_num > 2 else curv_ind
                data_dict_curv[curv_ind_j] = {'header': header_dict_top.copy() | header_dict.copy(), #merge header dictionaries
                                              'data': {},
                                              'units': {'x': header_dict['X axis unit [General Info]'],
                                                        'y': header_dict['Conversion Factor 00 [General Info]'].split(' ')[-1]
                                                        }
                                            } 
                for i, curv_dir in enumerate(line_order):
                    data_dict_curv[curv_ind_j]['data'][curv_dir] = {
                        'x': x_data[k+i],
                        'y': chan_offs+(y_data[k+i]*chan_fact)
                        }                                                  
            
            if curv_ind == curv_num:
                break
            else:
                pos += data_len #bytes read so far
                file.seek(pos, 0)

        data_dict[y_label] = {'image': data_dict_chan,
                            'curves': data_dict_curv
                            }
        file.close()
        
        return data_dict, y_label

    # read *.cur WSxM file
    @staticmethod
    def _wsxm_readcur(filepath):
        """
        Reads WSxM spectroscopy curve (*.cur) files and extracts the data.
        Args:
            filepath (str): The path to the .cur file to be read.
        Returns:
            tuple: A tuple containing:
                - data_dict (dict): A dictionary containing the extracted data.
                - y_label (str): The channel name for the Y-axis data.
        The structure of the returned data_dict is as follows:
        {
            'Y axis channel label': {
                'curves': {
                    curve_index: {
                        'header': header_dict,
                        'units': {'x': x_unit, 'y': y_unit},
                        'data': {
                            'approach': {'x': x_data, 'y': y_data},
                            'retract': {'x': x_data, 'y': y_data}
                },
                'image': {}
            }
        }
        Note:
            - The 'image' key is empty in this case as it does not exist in .cur files.
        """

        data_dict = {}
        file = open(f'{filepath}','rb')
        header_dict, pos = WSxMFuncs._wsxm_readheader(file)
        header_dict['File path'] = filepath #file path included to header

        if 'Index of this Curve [Control]' in header_dict.keys(): #for spectroscopy curves
            line_num = int(header_dict['Number of lines [General Info]'])
            y_label = header_dict['Y axis text [General Info]'].split('[')[0].strip()
            x_label = header_dict['X axis text [General Info]'].split('[')[0].strip()
            if header_dict['Index of this Curve [Control]'] == 'Average': #for average curves
                curv_ind = header_dict['Index of this Curve [Control]']
            else:
                curv_ind = int(header_dict['Index of this Curve [Control]'])
            curv_num = int(header_dict['Number of Curves in this serie [Control]'])
            chan_fact = float(header_dict['Conversion Factor 00 [General Info]'].split(' ')[0])
            y_unit = header_dict['Conversion Factor 00 [General Info]'].split(' ')[-1]
            chan_inv = header_dict['Channel is inverted [General Info]']
            if chan_inv == 'Yes':
                chan_fact = -chan_fact
            chan_offs = float(header_dict['Conversion Offset 00 [General Info]'].split(' ')[0])
            
            line_order = ['approach', 'retract']
            if header_dict['First Forward [Miscellaneous]'] == 'No': #CHECK THIS
                line_order = ['retract', 'approach']
        else: #for other kinds of *.cur (e.g. tune data)
            line_pts = int(header_dict['Number of points [General Info]'])
            line_num = int(header_dict['Number of lines [General Info]'])
            y_label = header_dict['Y axis text [General Info]'].split('[')[0].strip()
            x_label = header_dict['X axis text [General Info]'].split('[')[0].strip()
            #set generic values for irrelevant parameters here
            curv_ind = 1
            curv_num = 1
            chan_fact = 1
            chan_offs = 0  
            y_unit = header_dict['Y axis unit [General Info]']                         
            line_order = [f'{y_label}_{ln_i+1}' for ln_i in range(line_num)]
        
        header_dict['Spectroscopy channel'] = y_label #Insert channel name information into dictionary
        file.seek(pos, 0)
        data = file.read()
        data_list = []
        for ln in data.splitlines():
            ln_array = ln.decode('latin-1', errors='ignore').strip().replace('#QNAN','').split(' ')
            data_list.append(list(map(float,ln_array)))
        data_mat = np.array(data_list) #data matrix   

        if y_label not in data_dict.keys():
            data_dict[y_label] = {'curves':{}, 'image':{}}

        for j in range(int(line_num/len(line_order))):
            k = 2*len(line_order) * j
            curv_ind_j = f'{curv_ind}{chr(ord("a")+j)}' if line_num > 2 else curv_ind
            data_dict[y_label]['curves'][curv_ind_j] = {'header': header_dict.copy(), 
                                                        'data': {},
                                                        'units': {'x': header_dict['X axis unit [General Info]'],
                                                                'y': y_unit
                                                                }
                                                        }
            for i, curv_dir in enumerate(line_order):
                data_dict[y_label]['curves'][curv_ind_j]['data'][curv_dir] = {
                    'x': data_mat[:,k+(2*i)],
                    'y': chan_offs+(data_mat[:,k+(2*i+1)]*chan_fact) #converted to units
                    }

        file.close()
        
        return data_dict, y_label

    #read *.stp spectroscopy curves. Pass "data_dict" to update data of both approach and retract into it correctly
    @staticmethod
    def _wsxm_readstp(filepath, data_dict={}):
        """
        Reads a WSxM .stp spectroscopy file and extracts the data into a dictionary.
        Args:
            filepath (str): The path to the .stp file to be read.
            data_dict (dict, optional): A dictionary to store the extracted data. 
            Use this to combine "approach" and "retract" curve data spread across different files into the same dictionary passed here.
            Defaults to an empty dictionary.
        Returns:
            tuple: A tuple containing:
                - data_dict (dict): The dictionary containing the extracted data.
                - chan_label (str): The label of the spectroscopy channel.
        The structure of the returned data_dict is as follows:
        {
            'Channel label': {
                'curves': {
                    curve_index: {
                        'header': header_dict,
                        'units': {'x': x_unit', 'y': y_unit},
                        'data': {
                            'approach': {'x': x_data, 'y': y_data},
                            'retract': {'x': x_data, 'y': y_data}
                        }
                    }
                },
                'image': {}
            }
        }
        """

        file = open(f'{filepath}','rb')
        filename = filepath.name
        header_dict, pos = WSxMFuncs._wsxm_readheader(file)
        header_dict['File path'] = [filepath] #file path included to header
        data_format = header_dict['Image Data Type [General Info]']
        
        x_num = int(header_dict['Number of rows [General Info]'])
        y_num = int(header_dict['Number of columns [General Info]'])
        x_len = float(header_dict['X Amplitude [Control]'].split(' ')[0])
        y_len = float(header_dict['Y Amplitude [Control]'].split(' ')[0])
        z_len = float(header_dict['Z Amplitude [General Info]'].split(' ')[0])
        x_dir = header_dict['X scanning direction [General Info]']
        y_dir = header_dict['Y scanning direction [General Info]']
        file_dirkey = filename.split('.')[-2]
        if len(file_dirkey) == 1:
            chan_label = filename.split('_')[-1].split('.')[0] 
            z_dir = WSxMFuncs.SPECT_DICT[filename.split('.')[-2]]
        else:
            file_dirkey_match = re.search(r'line\_\d{1}', file_dirkey)
            chan_label = file_dirkey[:file_dirkey_match.start()].split('_')[-1]
            z_dir = WSxMFuncs.SPECT_DICT[x_dir] 
            
        if 'X starting offset [General Info]' in header_dict.keys(): #for "3D mode" images
            x_offset = float(header_dict['X starting offset [General Info]'].split(' ')[0])
            y_offset = float(header_dict['Y starting offset [General Info]'].split(' ')[0])
        else:
            x_offset = 0
            y_offset = 0

        header_dict['Spectroscopy channel'] = chan_label #Insert channel name information into dictionary

        z_data = np.linspace(x_offset, x_offset+x_len, y_num, endpoint=True) #CHECK THIS
        #read binary image data
        point_length, type_code  = WSxMFuncs.DATA_TYPES[data_format]
        file.seek(pos, 0)
        data_len = x_num*y_num*point_length
        bin_data = file.read(data_len)
        ch_array = np.array(list(struct.iter_unpack(f'{type_code}', bin_data))).flatten() 
        ch_mat = ch_array.reshape(x_num,y_num)
        if z_len == 0: #for zero data
            z_calib = 1
        else:
            z_calib = z_len/(ch_array.max()-ch_array.min())
        
        #create separate curve data for each line (consistent with '1D' data format)
        for i in range(x_num): 
            curv_ind = i + 1        
            #data dictionary initialised in a consistant format (also check wsxm_readcurves())
            if chan_label not in data_dict.keys():
                data_dict[chan_label] = {'curves': {}, 'image':{}}
            if curv_ind not in data_dict[chan_label]['curves'].keys():
                data_dict[chan_label]['curves'][curv_ind] = {'data': {},
                                                             'header': header_dict.copy(),
                                                             'units': {'x': header_dict['X Amplitude [Control]'].split(' ')[-1],
                                                                       'y': header_dict['Z Amplitude [General Info]'].split(' ')[-1]
                                                                       }
                                                            }
                #insert curve number info into header
                data_dict[chan_label]['curves'][curv_ind]['header']['Index of this Curve [Control]'] = str(curv_ind) 
                data_dict[chan_label]['curves'][curv_ind]['header']['Number of Curves in this serie [Control]'] = str(x_num)
            if z_dir not in data_dict[chan_label]['curves'][curv_ind]['data'].keys():
                data_dict[chan_label]['curves'][curv_ind]['data'][z_dir] = {}
            if filepath not in data_dict[chan_label]['curves'][curv_ind]['header']['File path']: #file path included to header
                data_dict[chan_label]['curves'][curv_ind]['header']['File path'].append(filepath)
            data_dict[chan_label]['curves'][curv_ind]['data'][z_dir]['x'] = z_data
            data_dict[chan_label]['curves'][curv_ind]['data'][z_dir]['y'] = (z_calib*ch_mat[:][i])
            if x_dir == 'Forward':
                data_dict[chan_label]['curves'][curv_ind]['data'][z_dir]['y'] = np.flip((z_calib*ch_mat[:][i]))

        file.close()
        
        return data_dict, chan_label
    

    def _wsxm_readforcevol(filepath):
        """
        Reads a WSxM .gsi force-volume data and extracts the data into a dictionary.
        Args:
            filepath (str): The path to the .stp file to be read.
        Returns:
            tuple: A tuple containing:
                - data_dict_chan (dict): The dictionary containing the extracted data.
                - chan_label (str): The label of the spectroscopy channel.
                - topo_data (numpy.ndarray): 2D array containing topography data.
        The structure of the returned data_dict_chan is as follows:
        {
            'header': header_dict,
            'units': {
                    'ZZ': data_unit,
                    'X': x_unit,
                    'Y': y_unit,
                    'Z': z_unit,
                    }
            'data': {
                    'ZZ': 3D data,
                    'X': x_data,
                    'Y': y_data,
                    'Z': z_data
                    },
            
        }
        """

        file = open(f'{filepath}','rb')
        header_dict, pos = WSxMFuncs._wsxm_readheader(file)
        header_dict['File path'] = filepath #file path included to header

        data_format = header_dict['Image Data Type [General Info]']
        chan_label = header_dict['Acquisition channel [General Info]']
        x_num = int(header_dict['Number of rows [General Info]'])
        y_num = int(header_dict['Number of columns [General Info]'])
        chan_num = int(header_dict['Number of points per ramp [General Info]'])
        x_len = float(header_dict['X Amplitude [Control]'].split(' ')[0])
        y_len = float(header_dict['Y Amplitude [Control]'].split(' ')[0])
        z_len = float(header_dict['Z Amplitude [General Info]'].split(' ')[0])
        chan_adc2v = float(header_dict['ADC to V conversion factor [General Info]'].split(' ')[0])
        chan_fact = float(header_dict['Conversion factor 0 for input channel [General Info]'].split(' ')[0])
        chan_offs = float(header_dict['Conversion offset 0 for input channel [General Info]'].split(' ')[0]) #0

        chan_inv = header_dict['Channel is inverted [General Info]']
        if chan_inv == 'Yes':
            chan_fact = -chan_fact
                
        x_data = np.linspace(0, x_len, x_num, endpoint=True)
        y_data = np.linspace(y_len, 0, y_num, endpoint=True)
    
        z_data = np.empty(0)
        for i in range(chan_num):
            z_data = np.append(z_data, float(header_dict[f'Image {i:03} [Spectroscopy images ramp value list]'].split(' ')[0]))

        z_data = np.flip(z_data) #reverse z data order to make zero as point of contact
        
        #read binary image data
        point_length, type_code  = WSxMFuncs.DATA_TYPES[data_format]
        file.seek(pos, 0)
        data_len = x_num*y_num*point_length
        #read first topography data
        bin_data = file.read(data_len)
        topo_array = np.array(list(struct.iter_unpack(f'{type_code}', bin_data))).flatten()
        if z_len == 0: #for zero data
            topo_calib = 1
        else:
            topo_calib = z_len/(topo_array.max()-topo_array.min())
        
        topo_data = topo_calib*topo_array.reshape(x_num, y_num) #topography data

        pos += data_len
        ch_array = np.empty(0) #initialize channel data array
        for i in range(1, chan_num+1):
            file.seek(pos, 0)
            bin_data = file.read(data_len)
            ch_array_temp = np.array(list(struct.iter_unpack(f'{type_code}', bin_data))).flatten()
            ch_array = np.append(ch_array, chan_offs+(ch_array_temp*chan_adc2v*chan_fact))
            pos += data_len #next image
            
        #img data dictionary
        data_dict_chan = {'data': {'ZZ': ch_array.reshape(chan_num,y_num,x_num),#,
                                'X': x_data,
                                'Y': y_data,
                                'Z': z_data
                                },
                        'header': header_dict,
                        'units': {'ZZ': header_dict['Conversion factor 0 for input channel [General Info]'].split(' ')[-1],
                                    'X': header_dict['X Amplitude [Control]'].split(' ')[-1],
                                    'Y': header_dict['Y Amplitude [Control]'].split(' ')[-1],
                                    'Z': header_dict['Z Amplitude [General Info]'].split(' ')[-1],
                                    }
                        }
        
        file.close()
        
        return data_dict_chan, chan_label, topo_data

    def _wsxm_readmovie(filepath):
        """
        Reads a WSxM .MOV and .mpp movie data and extracts the data into a dictionary.
        Args:
            filepath (str): The path to the .stp file to be read.
        Returns:
            tuple: A tuple containing:
                - data_dict_chan (dict): The dictionary containing the extracted data.
                - chan_label (str): The label of the spectroscopy channel.
        The structure of the returned data_dict_chan is as follows:
        {
            'header': header_dict,
            'units': {
                    'ZZ': data_unit,
                    'X': x_unit,
                    'Y': y_unit,
                    'Z': z_unit,
                    }
            'data': {
                    'ZZ': 3D data,
                    'X': x_data,
                    'Y': y_data,
                    'Z': z_data
                    },
            
        }
        """

        file = open(f'{filepath}','rb')
        header_dict, pos = WSxMFuncs._wsxm_readheader(file)
        header_dict['File path'] = filepath #file path included to header

        data_format = header_dict['Image Data Type [General Info]']
        chan_label = header_dict['Acquisition channel [General Info]']
        x_num = int(header_dict['Number of rows [General Info]'])
        y_num = int(header_dict['Number of columns [General Info]'])
        chan_num = int(header_dict['Number of Frames [General Info]'])
        x_len = float(header_dict['X Amplitude [Control]'].split(' ')[0])
        y_len = float(header_dict['Y Amplitude [Control]'].split(' ')[0])
        z_len = float(header_dict['Z Amplitude [General Info]'].split(' ')[0])
        z_min = float(header_dict['Minimum [Miscellaneous]'])
        z_max = float(header_dict['Maximum [Miscellaneous]'])
                
        x_data = np.linspace(0, x_len, x_num, endpoint=True)
        y_data = np.linspace(y_len, 0, y_num, endpoint=True)
        z_data = np.linspace(0, chan_num, chan_num, endpoint=True) #frame number array

        #read binary image data
        point_length, type_code  = WSxMFuncs.DATA_TYPES[data_format]
        data_len = x_num*y_num*point_length

        if z_len == 0: #for zero data
            z_calib = 1
        else:
            z_calib = z_len/(z_max-z_min)
        
        z_unit = header_dict['Z Amplitude [General Info]'].split(' ')[-1]
        #the following fixes a bug in the data format for amplitude channel, ensures that data is read in volts, not nanometers (which is fake)
        if chan_label == 'Amplitude' and header_dict['Z Amplitude [General Info]'].split(' ')[1] != 'V':
            chan_factor = float(header_dict['Conversion Factor 00 [General Info]'].split(' ')[0])
            z_calib = z_calib/chan_factor
            z_unit = 'V'

        ch_array = np.empty(0) #initialize channel data array
        for i in range(1, chan_num+1):
            file.seek(pos, 0)
            bin_data = file.read(data_len)
            ch_array_temp = np.array(list(struct.iter_unpack(f'{type_code}', bin_data))).flatten()          
            ch_array = np.append(ch_array, z_calib*ch_array_temp)
            pos += data_len #next image
            
        #img data dictionary
        data_dict_chan = {'data': {'ZZ': ch_array.reshape(chan_num,y_num,x_num),#,
                                'X': x_data,
                                'Y': y_data,
                                'Z': z_data
                                },
                        'header': header_dict,
                        'units': {'ZZ': z_unit,
                                    'X': header_dict['X Amplitude [Control]'].split(' ')[-1],
                                    'Y': header_dict['Y Amplitude [Control]'].split(' ')[-1],
                                    'Z': 'frame',
                                    }
                        }
        
        file.close()
        
        return data_dict_chan, chan_label



# -*- coding: utf-8 -*-
"""
Reader capable of reading one or all NSID datasets present in a given HDF5 file

Created on Fri May 22 16:29:25 2020

@author: Gerd Duscher, Suhas Somnath, Maxim Ziadtinov
"""


import h5py
import sidpy

from .hdf_utils import check_if_main, get_all_main, read_h5py_dataset


class NSIDReader(sidpy.Reader):

    def __init__(self, file_path):
        """
        Creates an instance of NSIDReader which can read one or more HDF5
        datasets formatted according to NSID into sidpy.Dataset objects

        Parameters
        ----------
        file_path : str, h5py.File, or h5py.Group
            Path to a HDF5 file or a handle to an open HDF5 file or group
            object

        Notes
        -----
        Please consider using the ``self._h5_file`` object to get handles to
        specific datasets or sub-trees that need to be read instead of opening
        the file again outside the context of this Reader.
        """
        super(NSIDReader, self).__init__(file_path)

        if not check_if_main:
            raise ModuleNotFoundError('Please install pyNSID to use this Reader')

        # Let h5py raise an OS error if a non-HDF5 file was provided
        self._h5_file = h5py.File(file_path, mode='r+')

        self._main_dsets = get_all_main(self._h5_file, verbose=False)

        # DO NOT close HDF5 file. Dask array will fail if you do so.

    def can_read(self):
        """
        Checks whether or not this Reader can read the provided file

        Returns
        -------
        bool :
            True if this Reader can read the provided file and if this file
            contains at least one NSID-formatted main dataset. Else, False
        """
        return len(self._main_dsets) > 0

    def read(self, h5_object=None):
        """
        Reads all available NSID main datasets or the specified h5_object

        Parameters
        ----------
        h5_object : h5py.Dataset or h5py.Group
            HDF5 Dataset to read or the HDF5 group under which to read all
            datasets

        Returns
        -------
        dictionary of sidpy.Dataset objects
            Datasets present in the provided file
        """
        if h5_object is None:
            return self.read_all(recursive=True)
        if not isinstance(h5_object, (h5py.Group, h5py.Dataset)):
            raise TypeError('Provided h5_object was not a h5py.Dataset or '
                            'h5py.Group object but was of type: {}'
                            ''.format(type(h5_object)))
        self.__validate_obj_in_same_file(h5_object)
        if isinstance(h5_object, h5py.Dataset):
            return read_h5py_dataset(h5_object)
        else:
            return self.read_all(parent=h5_object)
        
        #TODO: We need to add the ability to read functions in any sidpy datasets.
        # We will assume that fitting functions are contained within a dictionary of the same name. 
        #They will have been saved with dill and encoded. To decode them, use:

        '''
        fit_fn_packed = fit_data.metadata['fitting_functions']
        loaded_dict = {}
        fit_fns = []
        for key in fit_fn_packed.keys():
            encoded_value = fit_fn_packed['my_function']
            serialized_value = base64.b64decode(encoded_value)
            loaded_dict[key] = dill.loads(serialized_value)
            fit_fns.append(loaded_dict[key]) #retrieve the function
        '''

    def __validate_obj_in_same_file(self, h5_object):
        """
        Internal function that ensures that the provided HDF5 object is within
        the same file as that provided in __init__

        Parameters
        ----------
        h5_object : h5py.Dataset, h5py.Group
            HDF5 object

        Raises
        ------
        OSError - if the provded object is in a different HDF5 file.
        """
        if h5_object.file != self._h5_file:
            raise OSError('The file containing the provided h5_object: {} is '
                          'not the same as provided HDF5 file when '
                          'instantiating this object: {}'
                          ''.format(h5_object.file.filename,
                                    self._h5_file.filename))

    def read_all(self, recursive=True, parent=None):
        """
        Reads all HDF5 datasets formatted according to NSID specifications.

        Parameters
        ----------
        recursive : bool, default = True
            We might just remove this kwarg
        parent : h5py.Group, Default = None
            HDF5 group under which to read all available datasets.
            By default, all datasets within the HDF5 file are read.

        Returns
        -------
        dictionary of sidpy.Dataset objects
            Datasets present in the provided file
        """

        if parent is None:
            h5_group = self._h5_file
        else:
            if not isinstance(parent, h5py.Group):
                raise TypeError('parent should be a h5py.Group object')
            self.__validate_obj_in_same_file(parent)
            h5_group = parent

        if recursive:
            list_of_main = self._main_dsets
            keys = ['Channel_{:03}'.format(i) for i in range(len(list_of_main))]
        else:
            list_of_main = []
            keys = []
            for key in h5_group:
                if isinstance(h5_group[key], h5py.Dataset):
                    if check_if_main(h5_group[key]):
                        list_of_main.append(h5_group[key])
                        keys.append(key)

        # Go through each of the identified
        dictionary_of_datasets = {}
        for i,dset in enumerate(list_of_main):
            dictionary_of_datasets[keys[i]] = read_h5py_dataset(dset)
        return dictionary_of_datasets

    def close(self):
        self._h5_file.close()


"""
Tools to read, data in from generic files

Submodules
----------

.. autosummary::
    :toctree: _autosummary

    ImageReader
"""

from .hyperspy import convert_hyperspy

__all__ = ['convert_hyperspy']

all_readers = [convert_hyperspy]
